from __future__ import generator_stop

from ctypes import LittleEndianStructure
from ctypes.wintypes import ULONG, USHORT

from cwinsdk.shared.ntdef import UCHAR

# Windows Kits\10\Include\10.0.17763.0\km\ata.h


class GeneralConfiguration(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("Reserved1", USHORT, 1),
        ("Retired3", USHORT, 1),
        ("ResponseIncomplete", USHORT, 1),
        ("Retired2", USHORT, 3),
        ("FixedDevice", USHORT, 1),
        ("RemovableMedia", USHORT, 1),
        ("Retired1", USHORT, 7),
        ("DeviceType", USHORT, 1),
    ]


class TrustedComputing(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("FeatureSupported", USHORT, 1),
        ("Reserved", USHORT, 15),
    ]


class Capabilities(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("CurrentLongPhysicalSectorAlignment", UCHAR, 2),
        ("ReservedByte49", UCHAR, 6),
        ("DmaSupported", UCHAR, 1),
        ("LbaSupported", UCHAR, 1),
        ("IordyDisable", UCHAR, 1),
        ("IordySupported", UCHAR, 1),
        ("Reserved1", UCHAR, 1),
        ("StandybyTimerSupport", UCHAR, 1),
        ("Reserved2", UCHAR, 2),
        ("ReservedWord50", USHORT),
    ]


class AdditionalSupported(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("ZonedCapabilities", USHORT, 2),
        ("NonVolatileWriteCache", USHORT, 1),
        ("ExtendedUserAddressableSectorsSupported", USHORT, 1),
        ("DeviceEncryptsAllUserData", USHORT, 1),
        ("ReadZeroAfterTrimSupported", USHORT, 1),
        ("Optional28BitCommandsSupported", USHORT, 1),
        ("IEEE1667", USHORT, 1),
        ("DownloadMicrocodeDmaSupported", USHORT, 1),
        ("SetMaxSetPasswordUnlockDmaSupported", USHORT, 1),
        ("WriteBufferDmaSupported", USHORT, 1),
        ("ReadBufferDmaSupported", USHORT, 1),
        ("DeviceConfigIdentifySetDmaSupported", USHORT, 1),
        ("LPSAERCSupported", USHORT, 1),
        ("DeterministicReadAfterTrimSupported", USHORT, 1),
        ("CFastSpecSupported", USHORT, 1),
    ]


class SerialAtaCapabilities(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("Reserved0", USHORT, 1),  # shall be set to 0
        ("SataGen1", USHORT, 1),
        ("SataGen2", USHORT, 1),
        ("SataGen3", USHORT, 1),
        ("Reserved1", USHORT, 4),
        ("NCQ", USHORT, 1),
        ("HIPM", USHORT, 1),
        ("PhyEvents", USHORT, 1),
        ("NcqUnload", USHORT, 1),
        ("NcqPriority", USHORT, 1),
        ("HostAutoPS", USHORT, 1),
        ("DeviceAutoPS", USHORT, 1),
        ("ReadLogDMA", USHORT, 1),
        ("Reserved2", USHORT, 1),  # shall be set to 0
        ("CurrentSpeed", USHORT, 3),
        ("NcqStreaming", USHORT, 1),
        ("NcqQueueMgmt", USHORT, 1),
        ("NcqReceiveSend", USHORT, 1),
        ("DEVSLPtoReducedPwrState", USHORT, 1),
        ("Reserved3", USHORT, 8),
    ]


class SerialAtaFeaturesSupported(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("Reserved0", USHORT, 1),  # shall be set to 0
        ("NonZeroOffsets", USHORT, 1),
        ("DmaSetupAutoActivate", USHORT, 1),
        ("DIPM", USHORT, 1),
        ("InOrderData", USHORT, 1),
        ("HardwareFeatureControl", USHORT, 1),
        ("SoftwareSettingsPreservation", USHORT, 1),
        ("NCQAutosense", USHORT, 1),
        ("DEVSLP", USHORT, 1),
        ("HybridInformation", USHORT, 1),
        ("Reserved1", USHORT, 6),
    ]


class SerialAtaFeaturesEnabled(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("Reserved0", USHORT, 1),  # shall be set to 0
        ("NonZeroOffsets", USHORT, 1),
        ("DmaSetupAutoActivate", USHORT, 1),
        ("DIPM", USHORT, 1),
        ("InOrderData", USHORT, 1),
        ("HardwareFeatureControl", USHORT, 1),
        ("SoftwareSettingsPreservation", USHORT, 1),
        ("DeviceAutoPS", USHORT, 1),
        ("DEVSLP", USHORT, 1),
        ("HybridInformation", USHORT, 1),
        ("Reserved1", USHORT, 6),
    ]


class CommandSetSupport(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("SmartCommands", USHORT, 1),
        ("SecurityMode", USHORT, 1),
        ("RemovableMediaFeature", USHORT, 1),
        ("PowerManagement", USHORT, 1),  # shall be set to 1
        ("Reserved1", USHORT, 1),
        ("WriteCache", USHORT, 1),
        ("LookAhead", USHORT, 1),
        ("ReleaseInterrupt", USHORT, 1),
        ("ServiceInterrupt", USHORT, 1),
        ("DeviceReset", USHORT, 1),
        ("HostProtectedArea", USHORT, 1),
        ("Obsolete1", USHORT, 1),
        ("WriteBuffer", USHORT, 1),
        ("ReadBuffer", USHORT, 1),
        ("Nop", USHORT, 1),
        ("Obsolete2", USHORT, 1),
        ("DownloadMicrocode", USHORT, 1),
        ("DmaQueued", USHORT, 1),
        ("Cfa", USHORT, 1),
        ("AdvancedPm", USHORT, 1),
        ("Msn", USHORT, 1),
        ("PowerUpInStandby", USHORT, 1),
        ("ManualPowerUp", USHORT, 1),
        ("Reserved2", USHORT, 1),
        ("SetMax", USHORT, 1),
        ("Acoustics", USHORT, 1),
        ("BigLba", USHORT, 1),
        ("DeviceConfigOverlay", USHORT, 1),
        ("FlushCache", USHORT, 1),
        ("FlushCacheExt", USHORT, 1),
        ("WordValid83", USHORT, 2),  # shall be 01b
        ("SmartErrorLog", USHORT, 1),
        ("SmartSelfTest", USHORT, 1),
        ("MediaSerialNumber", USHORT, 1),
        ("MediaCardPassThrough", USHORT, 1),
        ("StreamingFeature", USHORT, 1),
        ("GpLogging", USHORT, 1),
        ("WriteFua", USHORT, 1),
        ("WriteQueuedFua", USHORT, 1),
        ("WWN64Bit", USHORT, 1),
        ("URGReadStream", USHORT, 1),
        ("URGWriteStream", USHORT, 1),
        ("ReservedForTechReport", USHORT, 2),
        ("IdleWithUnloadFeature", USHORT, 1),
        ("WordValid", USHORT, 2),  # shall be 01b
    ]


class CommandSetActive(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("SmartCommands", USHORT, 1),
        ("SecurityMode", USHORT, 1),
        ("RemovableMediaFeature", USHORT, 1),
        ("PowerManagement", USHORT, 1),
        ("Reserved1", USHORT, 1),
        ("WriteCache", USHORT, 1),
        ("LookAhead", USHORT, 1),
        ("ReleaseInterrupt", USHORT, 1),
        ("ServiceInterrupt", USHORT, 1),
        ("DeviceReset", USHORT, 1),
        ("HostProtectedArea", USHORT, 1),
        ("Obsolete1", USHORT, 1),
        ("WriteBuffer", USHORT, 1),
        ("ReadBuffer", USHORT, 1),
        ("Nop", USHORT, 1),
        ("Obsolete2", USHORT, 1),
        ("DownloadMicrocode", USHORT, 1),
        ("DmaQueued", USHORT, 1),
        ("Cfa", USHORT, 1),
        ("AdvancedPm", USHORT, 1),
        ("Msn", USHORT, 1),
        ("PowerUpInStandby", USHORT, 1),
        ("ManualPowerUp", USHORT, 1),
        ("Reserved2", USHORT, 1),
        ("SetMax", USHORT, 1),
        ("Acoustics", USHORT, 1),
        ("BigLba", USHORT, 1),
        ("DeviceConfigOverlay", USHORT, 1),
        ("FlushCache", USHORT, 1),
        ("FlushCacheExt", USHORT, 1),
        ("Resrved3", USHORT, 1),
        ("Words119_120Valid", USHORT, 1),
        ("SmartErrorLog", USHORT, 1),
        ("SmartSelfTest", USHORT, 1),
        ("MediaSerialNumber", USHORT, 1),
        ("MediaCardPassThrough", USHORT, 1),
        ("StreamingFeature", USHORT, 1),
        ("GpLogging", USHORT, 1),
        ("WriteFua", USHORT, 1),
        ("WriteQueuedFua", USHORT, 1),
        ("WWN64Bit", USHORT, 1),
        ("URGReadStream", USHORT, 1),
        ("URGWriteStream", USHORT, 1),
        ("ReservedForTechReport", USHORT, 2),
        ("IdleWithUnloadFeature", USHORT, 1),
        ("Reserved4", USHORT, 2),
    ]


class NormalSecurityEraseUnit(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("TimeRequired", USHORT, 15),
        ("ExtendedTimeReported", USHORT, 1),
    ]


class EnhancedSecurityEraseUnit(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("TimeRequired", USHORT, 15),
        ("ExtendedTimeReported", USHORT, 1),
    ]


class PhysicalLogicalSectorSize(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("LogicalSectorsPerPhysicalSector", USHORT, 4),
        ("Reserved0", USHORT, 8),
        ("LogicalSectorLongerThan256Words", USHORT, 1),
        ("MultipleLogicalSectorsPerPhysicalSector", USHORT, 1),
        ("Reserved1", USHORT, 2),
    ]


class CommandSetSupportExt(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("ReservedForDrqTechnicalReport", USHORT, 1),
        ("WriteReadVerify", USHORT, 1),
        ("WriteUncorrectableExt", USHORT, 1),
        ("ReadWriteLogDmaExt", USHORT, 1),
        ("DownloadMicrocodeMode3", USHORT, 1),
        ("FreefallControl", USHORT, 1),
        ("SenseDataReporting", USHORT, 1),
        ("ExtendedPowerConditions", USHORT, 1),
        ("Reserved0", USHORT, 6),
        ("WordValid", USHORT, 2),  # shall be 01b
    ]


class CommandSetActiveExt(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("ReservedForDrqTechnicalReport", USHORT, 1),
        ("WriteReadVerify", USHORT, 1),
        ("WriteUncorrectableExt", USHORT, 1),
        ("ReadWriteLogDmaExt", USHORT, 1),
        ("DownloadMicrocodeMode3", USHORT, 1),
        ("FreefallControl", USHORT, 1),
        ("SenseDataReporting", USHORT, 1),
        ("ExtendedPowerConditions", USHORT, 1),
        ("Reserved0", USHORT, 6),
        ("Reserved1", USHORT, 2),
    ]


class SecurityStatus(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("SecuritySupported", USHORT, 1),
        ("SecurityEnabled", USHORT, 1),
        ("SecurityLocked", USHORT, 1),
        ("SecurityFrozen", USHORT, 1),
        ("SecurityCountExpired", USHORT, 1),
        ("EnhancedSecurityEraseSupported", USHORT, 1),
        ("Reserved0", USHORT, 2),
        ("SecurityLevel", USHORT, 1),
        ("Reserved1", USHORT, 7),
    ]


class CfaPowerMode1(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("MaximumCurrentInMA", USHORT, 12),
        ("CfaPowerMode1Disabled", USHORT, 1),
        ("CfaPowerMode1Required", USHORT, 1),
        ("Reserved0", USHORT, 1),
        ("Word160Supported", USHORT, 1),
    ]


class DataSetManagementFeature(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("SupportsTrim", USHORT, 1),
        ("Reserved0", USHORT, 15),
    ]


class SCTCommandTransport(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("Supported", USHORT, 1),
        ("Reserved0", USHORT, 1),
        ("WriteSameSuported", USHORT, 1),
        ("ErrorRecoveryControlSupported", USHORT, 1),
        ("FeatureControlSuported", USHORT, 1),
        ("DataTablesSuported", USHORT, 1),
        ("Reserved1", USHORT, 6),
        ("VendorSpecific", USHORT, 4),
    ]


class BlockAlignment(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("AlignmentOfLogicalWithinPhysical", USHORT, 14),
        ("Word209Supported", USHORT, 1),  # shall be set to 1
        ("Reserved0", USHORT, 1),  # shall be cleared to 0
    ]


class NVCacheCapabilities(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("NVCachePowerModeEnabled", USHORT, 1),
        ("Reserved0", USHORT, 3),
        ("NVCacheFeatureSetEnabled", USHORT, 1),
        ("Reserved1", USHORT, 3),
        ("NVCachePowerModeVersion", USHORT, 4),
        ("NVCacheFeatureSetVersion", USHORT, 4),
    ]


class NVCacheOptions(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("NVCacheEstimatedTimeToSpinUpInSeconds", UCHAR),
        ("Reserved", UCHAR),
    ]


class TransportMajorVersion(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("MajorVersion", USHORT, 12),
        ("TransportType", USHORT, 4),
    ]


class IDENTIFY_DEVICE_DATA(LittleEndianStructure):
    __slots__ = ()
    _pack_ = 1
    _fields_ = [
        ("GeneralConfiguration", GeneralConfiguration),  # word 0
        ("NumCylinders", USHORT),
        ("SpecificConfiguration", USHORT),
        ("NumHeads", USHORT),
        ("Retired1", USHORT * 2),
        ("NumSectorsPerTrack", USHORT),
        ("VendorUnique1", USHORT * 3),
        ("SerialNumber", UCHAR * 20),
        ("Retired2", USHORT * 2),
        ("Obsolete1", USHORT),
        ("FirmwareRevision", UCHAR * 8),
        ("ModelNumber", UCHAR * 40),
        ("MaximumBlockTransfer", UCHAR),
        ("VendorUnique2", UCHAR),
        ("TrustedComputing", TrustedComputing),  # word 48
        ("Capabilities", Capabilities),
        ("ObsoleteWords51", USHORT * 2),
        ("TranslationFieldsValid", USHORT, 3),
        ("Reserved3", USHORT, 5),
        ("FreeFallControlSensitivity", USHORT, 8),
        ("NumberOfCurrentCylinders", USHORT),
        ("NumberOfCurrentHeads", USHORT),
        ("CurrentSectorsPerTrack", USHORT),  # word 56
        ("CurrentSectorCapacity", ULONG),
        ("CurrentMultiSectorSetting", UCHAR),
        ("MultiSectorSettingValid", UCHAR, 1),
        ("ReservedByte59", UCHAR, 3),
        ("SanitizeFeatureSupported", UCHAR, 1),
        ("CryptoScrambleExtCommandSupported", UCHAR, 1),
        ("OverwriteExtCommandSupported", UCHAR, 1),
        ("BlockEraseExtCommandSupported", UCHAR, 1),
        ("UserAddressableSectors", ULONG),
        ("ObsoleteWord62", USHORT),
        ("MultiWordDMASupport", USHORT, 8),
        ("MultiWordDMAActive", USHORT, 8),
        ("AdvancedPIOModes", USHORT, 8),
        ("ReservedByte64", USHORT, 8),
        ("MinimumMWXferCycleTime", USHORT),
        ("RecommendedMWXferCycleTime", USHORT),
        ("MinimumPIOCycleTime", USHORT),
        ("MinimumPIOCycleTimeIORDY", USHORT),
        ("AdditionalSupported", AdditionalSupported),
        ("ReservedWords70", USHORT * 5),  # word 74
        ("QueueDepth", USHORT, 5),
        ("ReservedWord75", USHORT, 11),
        ("SerialAtaCapabilities", SerialAtaCapabilities),
        ("SerialAtaFeaturesSupported", SerialAtaFeaturesSupported),
        ("SerialAtaFeaturesEnabled", SerialAtaFeaturesEnabled),
        ("MajorRevision", USHORT),
        ("MinorRevision", USHORT),
        ("CommandSetSupport", CommandSetSupport),
        ("CommandSetActive", CommandSetActive),
        ("UltraDMASupport", USHORT, 8),
        ("UltraDMAActive", USHORT, 8),
        ("NormalSecurityEraseUnit", NormalSecurityEraseUnit),
        ("EnhancedSecurityEraseUnit", EnhancedSecurityEraseUnit),  # word 90
        ("CurrentAPMLevel", USHORT, 8),
        ("ReservedWord91", USHORT, 8),
        ("MasterPasswordID", USHORT),
        ("HardwareResetResult", USHORT),
        ("CurrentAcousticValue", USHORT, 8),
        ("RecommendedAcousticValue", USHORT, 8),
        ("StreamMinRequestSize", USHORT),
        ("StreamingTransferTimeDMA", USHORT),
        ("StreamingAccessLatencyDMAPIO", USHORT),
        ("StreamingPerfGranularity", ULONG),
        ("Max48BitLBA", ULONG * 2),
        ("StreamingTransferTime", USHORT),
        ("DsmCap", USHORT),
        ("PhysicalLogicalSectorSize", PhysicalLogicalSectorSize),  # 106
        ("InterSeekDelay", USHORT),
        ("WorldWideName", USHORT * 4),
        ("ReservedForWorldWideName128", USHORT * 4),
        ("ReservedForTlcTechnicalReport", USHORT),
        ("WordsPerLogicalSector", USHORT * 2),
        ("CommandSetSupportExt", CommandSetSupportExt),
        ("CommandSetActiveExt", CommandSetActiveExt),
        ("ReservedForExpandedSupportandActive", USHORT * 6),  # 126
        ("MsnSupport", USHORT, 2),
        ("ReservedWord127", USHORT, 14),
        ("SecurityStatus", SecurityStatus),
        ("ReservedWord129", USHORT * 31),
        ("CfaPowerMode1", CfaPowerMode1),  # word 160
        ("ReservedForCfaWord161", USHORT * 7),
        ("NominalFormFactor", USHORT, 4),
        ("ReservedWord168", USHORT, 12),
        ("DataSetManagementFeature", DataSetManagementFeature),
        ("AdditionalProductID", USHORT * 4),
        ("ReservedForCfaWord174", USHORT * 2),
        ("CurrentMediaSerialNumber", USHORT * 30),
        ("SCTCommandTransport", SCTCommandTransport),  # 206
        ("ReservedWord207", USHORT * 2),
        ("BlockAlignment", BlockAlignment),
        ("WriteReadVerifySectorCountMode3Only", USHORT * 2),
        ("WriteReadVerifySectorCountMode2Only", USHORT * 2),
        ("NVCacheCapabilities", NVCacheCapabilities),
        ("NVCacheSizeLSW", USHORT),
        ("NVCacheSizeMSW", USHORT),
        ("NominalMediaRotationRate", USHORT),
        ("ReservedWord218", USHORT),
        ("NVCacheOptions", NVCacheOptions),
        ("WriteReadVerifySectorCountMode", USHORT, 8),
        ("ReservedWord220", USHORT, 8),
        ("ReservedWord221", USHORT),
        ("TransportMajorVersion", TransportMajorVersion),  # 222
        ("TransportMinorVersion", USHORT),
        ("ReservedWord224", USHORT * 6),
        ("ExtendedNumberOfUserAddressableSectors", ULONG * 2),
        ("MinBlocksPerDownloadMicrocodeMode03", USHORT),
        ("MaxBlocksPerDownloadMicrocodeMode03", USHORT),
        ("ReservedWord236", USHORT * 19),
        ("Signature", USHORT, 8),  # word 255
        ("CheckSum", USHORT, 8),
    ]


# SMART sub command list
IDE_SMART_READ_ATTRIBUTES = 0xD0
IDE_SMART_READ_THRESHOLDS = 0xD1
IDE_SMART_ENABLE_DISABLE_AUTOSAVE = 0xD2
IDE_SMART_SAVE_ATTRIBUTE_VALUES = 0xD3
IDE_SMART_EXECUTE_OFFLINE_DIAGS = 0xD4
IDE_SMART_READ_LOG = 0xD5
IDE_SMART_WRITE_LOG = 0xD6
IDE_SMART_ENABLE = 0xD8
IDE_SMART_DISABLE = 0xD9
IDE_SMART_RETURN_STATUS = 0xDA
IDE_SMART_ENABLE_DISABLE_AUTO_OFFLINE = 0xDB



from __future__ import generator_stop

import sys

if sys.platform != "win32":
    raise ImportError("win submodule is only available on Windows")

import re
from ctypes import Structure, WinError, byref, create_unicode_buffer, sizeof
from ctypes.wintypes import BYTE, DWORD
from typing import TYPE_CHECKING

from cwinsdk import struct2dict
from cwinsdk.um.fileapi import OPEN_EXISTING, CreateFileW
from cwinsdk.um.handleapi import INVALID_HANDLE_VALUE, CloseHandle
from cwinsdk.um.ioapiset import DeviceIoControl
from cwinsdk.um.winioctl import (
    DISK_GEOMETRY,
    FSCTL_ALLOW_EXTENDED_DASD_IO,
    FSCTL_LOCK_VOLUME,
    FSCTL_UNLOCK_VOLUME,
    GET_LENGTH_INFORMATION,
    GETVERSIONINPARAMS,
    IOCTL_DISK_GET_DRIVE_GEOMETRY,
    IOCTL_DISK_GET_LENGTH_INFO,
    IOCTL_DISK_VERIFY,
    IOCTL_STORAGE_QUERY_PROPERTY,
    IOCTL_STORAGE_READ_CAPACITY,
    PARTITION_IFS,
    PARTITION_MSFT_RECOVERY,
    SMART_GET_VERSION,
    STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR,
    STORAGE_DEVICE_DESCRIPTOR,
    STORAGE_PROPERTY_ID,
    STORAGE_PROPERTY_QUERY,
    STORAGE_QUERY_TYPE,
    STORAGE_READ_CAPACITY,
    VERIFY_INFORMATION,
)
from cwinsdk.um.winnt import FILE_SHARE_READ, FILE_SHARE_WRITE

from ..exceptions import assert_true
from .handle import WindowsHandle, _mode2access

if TYPE_CHECKING:
    from typing import Any


class EMPTY_BUFFER(Structure):
    pass


def get_length(path):
    # type: (str, ) -> int

    """Returns for:
    partitions: the full partition size (might be larger than disk_usage().total)
    drives: the full drive size (might be larger than the one computed from the drive geometry)
    """

    handle = open_device(path)
    try:
        InBuffer = EMPTY_BUFFER()
        OutBuffer = GET_LENGTH_INFORMATION()
        MyDeviceIoControl(handle, IOCTL_DISK_GET_LENGTH_INFO, InBuffer, OutBuffer)

        return OutBuffer.Length

    finally:
        CloseHandle(handle)


class Volume(WindowsHandle):
    def __init__(self, path):
        # type: (str, ) -> None

        handle = open_logical_volume(path)
        WindowsHandle.__init__(self, handle)

    def lock(self):
        MyDeviceIoControl(self.handle, FSCTL_LOCK_VOLUME, EMPTY_BUFFER(), EMPTY_BUFFER())

    def unlock(self):
        MyDeviceIoControl(self.handle, FSCTL_UNLOCK_VOLUME, EMPTY_BUFFER(), EMPTY_BUFFER())

    def extend(self):
        MyDeviceIoControl(self.handle, FSCTL_ALLOW_EXTENDED_DASD_IO, EMPTY_BUFFER(), EMPTY_BUFFER())


class Drive(WindowsHandle):
    def __init__(self, DriveIndex):
        # type: (int, ) -> None

        handle = open_physical_drive(DriveIndex)
        WindowsHandle.__init__(self, handle)

    def get_alignment(self):
        # type: () -> dict

        InBuffer = STORAGE_PROPERTY_QUERY()
        OutBuffer = STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR()

        InBuffer.PropertyId = STORAGE_PROPERTY_ID.StorageAccessAlignmentProperty
        InBuffer.QueryType = STORAGE_QUERY_TYPE.PropertyStandardQuery

        MyDeviceIoControl(self.handle, IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, OutBuffer)

        return struct2dict(OutBuffer)

    def get_capacity(self):
        # type: () -> dict

        """Accept drive and volume handles, but always returns related drive information.
        The DiskLength returned is the same as if `get_length` is used on the related drive.
        """

        InBuffer = EMPTY_BUFFER()
        OutBuffer = STORAGE_READ_CAPACITY()
        MyDeviceIoControl(self.handle, IOCTL_STORAGE_READ_CAPACITY, InBuffer, OutBuffer)

        return struct2dict(OutBuffer)

    # was: StorageQuery
    def get_device(self):
        # type: () -> dict

        InBuffer = STORAGE_PROPERTY_QUERY()
        OutBuffer = STORAGE_DEVICE_DESCRIPTOR()

        InBuffer.PropertyId = STORAGE_PROPERTY_ID.StorageDeviceProperty
        InBuffer.QueryType = STORAGE_QUERY_TYPE.PropertyStandardQuery
        InBuffer.AdditionalParameters = (BYTE * 1)(0)

        MyDeviceIoControl(self.handle, IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, OutBuffer)

        return struct2dict(OutBuffer)

    def get_geometry(self):

        InBuffer = EMPTY_BUFFER()
        OutBuffer = DISK_GEOMETRY()
        MyDeviceIoControl(self.handle, IOCTL_DISK_GET_DRIVE_GEOMETRY, InBuffer, OutBuffer)

        return struct2dict(OutBuffer)

    def get_smart_version(self):
        # type: () -> dict

        InBuffer = EMPTY_BUFFER()
        OutBuffer = GETVERSIONINPARAMS()

        MyDeviceIoControl(self.handle, SMART_GET_VERSION, InBuffer, OutBuffer)

        return struct2dict(OutBuffer)

    def verify_extent(self, start, length):
        # type: (int, int) -> None

        """Can cause issues with some drivers (like TrueCrypt)."""

        InBuffer = VERIFY_INFORMATION()
        OutBuffer = EMPTY_BUFFER()

        InBuffer.StartingOffset = start
        InBuffer.Length = length

        MyDeviceIoControl(self.handle, IOCTL_DISK_VERIFY, InBuffer, OutBuffer)


def get_drive_size(DriveIndex):
    with Drive(DriveIndex) as drive:
        geo = drive.get_geometry()
    return geo["Cylinders"] * geo["TracksPerCylinder"] * geo["SectorsPerTrack"] * geo["BytesPerSector"]


def partition_style_string(partition_style):
    return {0: "MBR", 1: "GPT"}.get(partition_style, "Unknown")


def partition_type_string(partition_type):
    return {PARTITION_IFS: "NTFS/exFAT", PARTITION_MSFT_RECOVERY: "Recovery"}.get(partition_type, "Unknown")


def MyDeviceIoControl(DeviceHandle, IoControlCode, InBuffer, OutBuffer, check_output=True):
    # type: (Any, Any, Any, Any, bool) -> None

    BytesReturned = DWORD()

    assert_true("IoControlCode", IoControlCode)

    ret = DeviceIoControl(
        DeviceHandle,
        IoControlCode,
        byref(InBuffer),
        sizeof(InBuffer),
        byref(OutBuffer),
        sizeof(OutBuffer),
        byref(BytesReturned),
        None,
    )

    if ret == 0:
        raise WinError()

    if check_output and BytesReturned.value != sizeof(OutBuffer):
        raise RuntimeError(f"DeviceIoControl expected {sizeof(OutBuffer)} bytes but got {BytesReturned.value} bytes")


def open_device(FileName, mode="r"):

    DesiredAccess = _mode2access[mode]

    DeviceHandle = CreateFileW(
        FileName, DesiredAccess, FILE_SHARE_READ | FILE_SHARE_WRITE, None, OPEN_EXISTING, 0, None
    )

    if DeviceHandle == INVALID_HANDLE_VALUE:
        raise WinError()

    return DeviceHandle


volumep = re.compile(r"^\\\\[\.\?]\\[A-Z]:$")
drivep = re.compile(r"^\\\\[\.\?]\\PHYSICALDRIVE[0-9]$")


def is_volume_or_drive(s):

    from argparse import ArgumentTypeError

    pre = "\\\\.\\"
    if not s.startswith(pre):
        s = pre + s

    s = s.upper()

    if volumep.match(s) or drivep.match(s):
        return s
    else:
        raise ArgumentTypeError("X: or PHYSICALDRIVEX")


def open_logical_volume(FileName):
    assert volumep.match(FileName)
    return open_device(FileName)


def open_physical_drive(DriveIndex):
    # type: (int, ) -> Any

    assert isinstance(DriveIndex, int)
    drive = fr"\\.\PHYSICALDRIVE{DriveIndex}"
    FileName = create_unicode_buffer(drive)

    return open_device(FileName)


if __name__ == "__main__":

    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("driveindex", type=int)
    args = parser.parse_args()

    with Drive(args.driveindex) as d:
        print("Alignment:", d.get_alignment())
        print("Capacity:", d.get_capacity())
        print("Device:", d.get_device())
        print("Geometry:", d.get_geometry())
        try:
            print("SMART version:", d.get_smart_version())
        except OSError as e:
            print(e)



from __future__ import generator_stop

from typing import TYPE_CHECKING

from pptx import Presentation  # python-pptx

if TYPE_CHECKING:
    from typing import Iterator, List


def read_ppt(path):
    # type: (str, ) -> Iterator[List[str]]

    prs = Presentation(path)
    for slide in prs.slides:
        shapes = [shape.text for shape in slide.shapes if hasattr(shape, "text")]
        yield shapes



from __future__ import generator_stop

from datetime import datetime
from functools import partial

from msgpack import ExtType, packb, unpackb

from .compat.datetime import datetime as compat_datetime


class EXTID:
    datetime = 1
    set = 2
    frozenset = 3


def default(obj):
    if isinstance(obj, (datetime, compat_datetime)):
        return ExtType(EXTID.datetime, obj.isoformat().encode("utf-8"))  # todo: use struct.Struct instead
    elif isinstance(obj, set):
        return ExtType(EXTID.set, packb(tuple(obj), use_bin_type=True))
    elif isinstance(obj, frozenset):
        return ExtType(EXTID.frozenset, packb(tuple(obj), use_bin_type=True))

    return obj


def ext_hook(code, data):
    if code == EXTID.datetime:
        return compat_datetime.fromisoformat(data.decode("utf-8"))
    elif code == EXTID.set:
        return set(unpackb(data, use_list=False, raw=False))
    elif code == EXTID.frozenset:
        return frozenset(unpackb(data, use_list=False, raw=False))

    return data


dumps = partial(packb, default=default, use_bin_type=True)
loads = partial(unpackb, ext_hook=ext_hook, use_list=False, raw=False)



from __future__ import generator_stop

from itertools import cycle
from typing import Iterable, Iterator


def xor_stream(it, key, cyclekey=True):
    # type: (Iterable[bytes], Iterable[bytes], bool) -> Iterator[bytes]

    if cyclekey:
        key = cycle(key)

    for i, j in zip(it, key):
        yield bytes([i ^ j])



from __future__ import generator_stop

import os.path
from os import fspath, remove, replace
from tempfile import mkstemp
from typing import TYPE_CHECKING

from .file import copen

if TYPE_CHECKING:
    from os import PathLike
    from typing import IO, ContextManager, Optional, Union

    PathType = Union[str, PathLike]

# http://stupidpythonideas.blogspot.tw/2014/07/getting-atomic-writes-right.html
class TransactionalCreateFile:
    def __init__(self, path, mode="wb", encoding=None, errors=None, newline=None, prefix="tmp"):
        # type: (PathType, str, Optional[str], Optional[str], Optional[str], str) -> None

        is_text = "t" in mode

        self.path = fspath(path)
        suffix = os.path.splitext(self.path)[1].lower()
        curdir = os.path.dirname(self.path)
        fd, self.tmppath = mkstemp(suffix, prefix, curdir, is_text)
        self.fp = copen(fd, mode, encoding=encoding, errors=errors, newline=newline, ext=suffix)

    def commit(self):
        # type: () -> None

        self.fp.close()
        replace(self.tmppath, self.path)  # should be atomic

    def rollback(self):
        # type: () -> None

        self.fp.close()
        remove(self.tmppath)

    def __enter__(self):
        # type: () -> IO

        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        # at this point the original file is unmodified and the new file exists as tempfile on disk (or in buffer on windows)
        if exc_type:
            self.rollback()
        else:
            self.commit()


def sopen(path, mode="rb", encoding=None, errors=None, newline=None, safe=False):
    # type: (PathType, str, Optional[str], Optional[str], Optional[str], bool) -> ContextManager[IO]

    if safe:
        return TransactionalCreateFile(path, mode, encoding=encoding, errors=errors, newline=newline)
    else:
        return copen(path, mode, encoding=encoding, errors=errors, newline=newline)


def write_file(data, path, mode="wb", encoding=None, errors=None, newline=None):
    # type: (Union[str, bytes], PathType, str, Optional[str], Optional[str], Optional[str]) -> None

    """Writes/overwrites files in a safe way. That means either the original file
    will be left untouched, or it will be replaced with the complete new file.
    """

    with TransactionalCreateFile(path, mode, encoding=encoding, errors=errors, newline=newline) as fw:
        fw.write(data)



from __future__ import generator_stop

from copy import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List


def cast(object, class_, instanceof=object, *args, **kwargs):
    """Changes the class of `object` to `class_` if `object` is an instance of `instanceof`,
    calls the initializer and returns it.
    """

    object = copy(object)
    if isinstance(object, instanceof):
        object.__class__ = class_
        object.__init__(*args, **kwargs)
    else:
        raise TypeError(f"Object is not an instance of {instanceof.__name__}")
    return object


class STAR:
    pass


def args_to_key(args, kwargs, separator=STAR):
    # type: (tuple, dict, Any) -> tuple

    """Create cache key from function arguments."""

    key = []  # type: List[tuple]
    if args:
        key.extend(args)
    if kwargs:
        key.append(separator)
        key.extend(sorted(kwargs.items()))

    return tuple(key)


def compress(value):
    # type: (Any, ) -> Any

    """Creates a copy of the object where some data structures are replaced with equivalent ones
    which take up less space, but are not necessarily mutable anymore.

    tuple < list
    set == frozenset
    bytes < bytearray
    """

    # sets are not processed because they cannot contain lists or bytearrays anyway.

    if isinstance(value, (tuple, list)):  # tuple *can* contain mutables
        return tuple(compress(x) for x in value)
    elif isinstance(value, bytearray):
        return bytes(value)  # bytearray can only be bytes or List[int] right?
    elif isinstance(value, dict):
        return {k: compress(v) for k, v in value.items()}
    else:
        return value



from __future__ import generator_stop

import logging
import os
import signal
from ctypes import WinError, byref, c_wchar_p, cast, create_unicode_buffer, memset, sizeof
from ctypes.wintypes import DWORD, LPCWSTR, ULARGE_INTEGER
from msvcrt import get_osfhandle
from os import fspath
from typing import TYPE_CHECKING

from cwinsdk.shared.ehstorioctl import MAX_PATH
from cwinsdk.shared.ntdef import PWSTR
from cwinsdk.um.combaseapi import CoTaskMemFree
from cwinsdk.um.consoleapi import ENABLE_VIRTUAL_TERMINAL_PROCESSING, GetConsoleMode, SetConsoleMode
from cwinsdk.um.fileapi import (
    INVALID_FILE_ATTRIBUTES,
    GetDiskFreeSpaceExW,
    GetFileAttributesW,
    GetVolumeInformationW,
    LockFileEx,
    UnlockFileEx,
)
from cwinsdk.um.KnownFolders import FOLDERID_RoamingAppData
from cwinsdk.um.minwinbase import LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY, OVERLAPPED
from cwinsdk.um.processenv import GetStdHandle
from cwinsdk.um.ShlObj_core import SHGetKnownFolderPath
from cwinsdk.um.WinBase import STD_OUTPUT_HANDLE
from cwinsdk.um.winnt import FILE_ATTRIBUTE_REPARSE_POINT

from .os_shared import _usagetuple, _volumeinfotuple

if TYPE_CHECKING:
    from ctypes.wintypes import HANDLE
    from os import PathLike
    from typing import IO, Union

    PathType = Union[str, PathLike]

unc_prefix = "\\\\?\\"


def get_stdout_handle():
    # type: () -> HANDLE

    """Might return a redirect handle. For the real handle, use CreateFile("CONOUT$")"""

    return GetStdHandle(STD_OUTPUT_HANDLE)


class EnableAnsi:  # doesn't work for some reason...
    def __init__(self):
        # type: () -> None

        # os.system("") # calls cmd, which sets ANSI mode, but doesn't disable it when exiting (it's a bug probably)
        self.handle = get_stdout_handle()
        self.oldmode = DWORD()

        GetConsoleMode(self.handle, byref(self.oldmode))
        SetConsoleMode(self.handle, self.oldmode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING)

    def close(self):
        # type: () -> None

        SetConsoleMode(self.handle, self.oldmode.value)

    def __enter__(self):
        return None

    def __exit__(self, *args):
        self.close()


def _islink(path):
    # type: (PathType, ) -> bool

    """Tests if `path` refers to a symlink or a junction.
    - Python >= 3.2 `os.path.islink()` only supports symlinks, not junctions.
    - Python < 3.2 `os.path.islink()` always returns `False` on Windows.
    This function works in all cases.
    """

    FileName = fspath(path)
    FileAttributes = GetFileAttributesW(FileName)

    if FileAttributes == INVALID_FILE_ATTRIBUTES:
        raise WinError()
    return FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT == FILE_ATTRIBUTE_REPARSE_POINT


def _uncabspath(path):
    # type: (str, )-> str

    if path.startswith(unc_prefix):
        return os.path.abspath(path)
    else:
        return unc_prefix + os.path.abspath(path)


def _lock(fp, exclusive=True, block=False):
    # type: (IO, bool, bool) -> None

    fd = fp.fileno()
    handle = get_osfhandle(fd)

    flags = 0
    if exclusive:
        flags |= LOCKFILE_EXCLUSIVE_LOCK
    if not block:
        flags |= LOCKFILE_FAIL_IMMEDIATELY

    overlapped = OVERLAPPED()
    memset(byref(overlapped), 0, sizeof(overlapped))

    LockFileEx(handle, flags, 0, 0xFFFFFFFF, 0xFFFFFFFF, overlapped)


def _unlock(fp):
    # type: (IO) -> None

    fd = fp.fileno()
    handle = get_osfhandle(fd)

    overlapped = OVERLAPPED()
    memset(byref(overlapped), 0, sizeof(overlapped))

    UnlockFileEx(handle, 0, 0xFFFFFFFF, 0xFFFFFFFF, overlapped)


def _get_appdata_dir():
    # type: () -> str

    KF_FLAG_CREATE = 0x00008000

    rfid = byref(FOLDERID_RoamingAppData)
    Flags = KF_FLAG_CREATE
    Token = None
    Path = PWSTR()

    try:
        result = SHGetKnownFolderPath(rfid, Flags, Token, byref(Path))
    except OSError:
        logging.error(f"SHGetKnownFolderPath result: {result & 0xffffffff:X}")
        raise

    ret = cast(Path, c_wchar_p).value
    CoTaskMemFree(Path)
    assert ret
    return ret


def _disk_usage_windows(path):
    # type: (str, ) -> _usagetuple

    DirectoryName = LPCWSTR(path)
    FreeBytesAvailableToCaller = ULARGE_INTEGER(0)  # user free
    TotalNumberOfBytes = ULARGE_INTEGER(0)  # user total
    TotalNumberOfFreeBytes = None  # total free

    GetDiskFreeSpaceExW(
        DirectoryName, byref(FreeBytesAvailableToCaller), byref(TotalNumberOfBytes), TotalNumberOfFreeBytes
    )
    return _usagetuple(
        TotalNumberOfBytes.value,
        TotalNumberOfBytes.value - FreeBytesAvailableToCaller.value,
        FreeBytesAvailableToCaller.value,
    )


def _volume_info_windows(path):
    # type: (str, ) -> _volumeinfotuple

    if not path.endswith("\\"):
        raise ValueError("X: usually doesn't work. X:\\ does.")

    RootPathName = LPCWSTR(path)
    VolumeNameBuffer = create_unicode_buffer(MAX_PATH + 1)
    VolumeNameSize = MAX_PATH + 1
    VolumeSerialNumber = DWORD()
    MaximumComponentLength = DWORD()
    FileSystemFlags = DWORD()
    FileSystemNameBuffer = create_unicode_buffer(MAX_PATH + 1)
    FileSystemNameSize = MAX_PATH + 1

    GetVolumeInformationW(
        RootPathName,
        VolumeNameBuffer,
        VolumeNameSize,
        byref(VolumeSerialNumber),
        byref(MaximumComponentLength),
        byref(FileSystemFlags),
        FileSystemNameBuffer,
        FileSystemNameSize,
    )

    return _volumeinfotuple(
        VolumeNameBuffer.value,
        VolumeSerialNumber.value,
        MaximumComponentLength.value,
        FileSystemFlags.value,
        FileSystemNameBuffer.value,
    )


def _interrupt_windows():
    # type: () -> None

    os.kill(os.getpid(), signal.CTRL_C_EVENT)  # fixme: verify: works on win 10 but not on win 7


def _filemanager_cmd_windows(path):
    # type: (str, ) -> str

    return f'explorer.exe /select,"{path}"'


if __name__ == "__main__":
    s = "\033[35m" + "color-test" + "\033[39m" + " test end"
    print(s)
    with EnableAnsi():
        print(s)
    print(s)



from __future__ import generator_stop

from itertools import count
from socket import timeout as SocketTimeout
from time import sleep

from hyper import HTTP20Connection
from hyper.http20.exceptions import StreamResetError

from .http import HTTPError


class WrappedHTTP20Connection:
    def __init__(self, host):
        self.host = host

    def connect(self):
        self.conn = HTTP20Connection(self.host)

    def _dl_url(self, path, headers, timeout):
        self.conn.request("GET", path, headers=headers)
        self.conn._sock._sck.settimeout(timeout)
        resp = self.conn.get_response()
        if resp.status >= 200 and resp.status < 300:
            return resp.read()

        print(path, resp.status, resp.trailers, resp.headers)
        raise HTTPError("non-2xx error", response=resp)

    def dl_url(self, path, headers=None, timeout=60):
        for i in count(1):
            sleep(i)
            try:
                return self._dl_url(path, headers, timeout)
            except (ConnectionResetError, SocketTimeout, StreamResetError):
                self.connect()
                continue



from __future__ import generator_stop

from random import choice, randrange
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator, Tuple


def randstr(length: int, charset: str) -> str:

    """Returns a (noncryptographic) random string consisting of characters from `charset`
    of length `length`.
    """

    return "".join(choice(charset) for i in range(length))  # nosec


def randbytes(size: int) -> bytes:

    """Returns (noncryptographic) random bytes of length `length`."""

    return bytes(randrange(0, 256) for _ in range(size))  # nosec


def rgb_colors():
    # type: () -> Iterator[Tuple[int, int, int]]

    """Yields a stream of (noncryptographic) random RGB color tuples."""

    while True:
        rgb = randrange(0, 256 ** 3)  # nosec
        rg, b = divmod(rgb, 256)
        r, g = divmod(rg, 256)
        yield (r, g, b)



from __future__ import generator_stop

from typing import TYPE_CHECKING

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_der_private_key,
    load_pem_private_key,
)
from OpenSSL import crypto  # pyOpenSSL

from .exceptions import assert_choice

if TYPE_CHECKING:
    from typing import Union

    from cryptography.hazmat.primitives.asymmetric.dh import DHPrivateKey
    from cryptography.hazmat.primitives.asymmetric.dsa import DSAPrivateKey
    from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
    from OpenSSL.crypto import X509

    PrivateKey = Union[RSAPrivateKey, DSAPrivateKey, DHPrivateKey, EllipticCurvePrivateKey]


def load_certificate_file(filename):
    # type: (str, ) -> X509

    with open(filename, "rb") as fr:
        return crypto.load_certificate(crypto.FILETYPE_PEM, fr.read())


def get_pubkey_from_x509(x509):
    # type: (X509, ) -> bytes

    return crypto.dump_privatekey(crypto.FILETYPE_ASN1, x509.get_pubkey())


def load_keyfile(filename, encoding="PEM"):
    # type: (str, str) -> PrivateKey

    try:
        load_func = {
            "PEM": load_pem_private_key,
            "DER": load_der_private_key,
        }[encoding]
    except KeyError:
        raise ValueError("Invalid encoding")

    with open(filename, "rb") as fr:
        password = None
        backend = default_backend()
        return load_func(fr.read(), password, backend)


def generate_rsa_keyfile_pair(priv_key, pub_key, key_size=4096, encoding="PEM", format="modern"):
    # type: (str, str, int, str, str) -> RSAPrivateKey

    """
    encoding: "PEM" -> ASN.1 encoding, "DER" -> base64 PEM
    """

    if key_size < 2048:
        raise ValueError("key_size must be atleast 2048")

    encodings = {
        "PEM": Encoding.PEM,
        "DER": Encoding.DER,
    }

    private_formats = {
        "modern": PrivateFormat.PKCS8,
        "legacy": PrivateFormat.TraditionalOpenSSL,
    }

    public_formats = {
        "modern": PublicFormat.SubjectPublicKeyInfo,
        "legacy": PublicFormat.PKCS1,
    }

    assert_choice("encoding", encoding, encodings)
    assert_choice("format", format, private_formats)

    public_exponent = 65537
    backend = default_backend()

    key = generate_private_key(public_exponent, key_size, backend)

    with open(priv_key, "wb") as fw:
        fw.write(key.private_bytes(encodings[encoding], private_formats[format], NoEncryption()))

    with open(pub_key, "wb") as fw:
        fw.write(key.public_key().public_bytes(encodings[encoding], public_formats[format]))

    return key



from __future__ import generator_stop

from collections import UserDict, defaultdict
from typing import Any, Callable, Dict, Hashable, Iterable, Iterator, List, Mapping, Tuple, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


def flatten(d: Union[Dict[T, U], List[U], Tuple[U, ...]]) -> Iterator[U]:

    """Flattens dicts/lists of (dicts/lists of ...) lists to lists."""

    if isinstance(d, (list, tuple)):
        yield from d
    elif isinstance(d, dict):
        for k, val in d.items():
            yield from flatten(val)
    else:
        raise TypeError(f"Unsupported type: {type(d)}")


def get_one_of(d: Dict[T, U], keys: Iterable[T]) -> Tuple[T, U]:

    """Returns the (key, value) pair of the first key of `keys` found in `d`."""

    for key in keys:
        try:
            return key, d[key]
        except KeyError:
            pass

    raise KeyError("None of the keys could be found")


# similar: subdict
def get_available(d: Dict[T, U], keys: Iterable[T]) -> Iterator[Tuple[T, U]]:

    """Returns all the key-value pairs in `d` for the keys in `it`.
    Missing keys are ignored.
    """

    for key in keys:
        try:
            yield key, d[key]
        except KeyError:
            pass


def subdict(d: Mapping[T, U], it: Iterable[T]) -> Dict[T, U]:

    """Uses the elements of `it` as keys to extract a new sub-dictionary.
    Raises if not all keys in `it` are available.
    """

    return {key: d[key] for key in it}


# was: mapdict, mapget
def mapmap(d: Mapping[T, U], it: Iterable[T]) -> Iterator[U]:
    """Returns all the values of `d` for the keys in `it`.
    Raises for missing keys.
    """

    return (d[i] for i in it)


def hasvalues(d: dict) -> dict:

    """Returns a sub-dictionary which leaves out all pairs where the value evaluates to False."""

    return {k: v for k, v in d.items() if v}


def valuemap(func: Callable[[U], V], d: Dict[T, U]) -> Dict[T, V]:

    """Returns a new dictionary with `func` applied to all values of `d`."""

    return {k: func(v) for k, v in d.items()}


def itemgetter(it: Iterable[T]) -> Callable[[Mapping[T, U]], Iterator[U]]:

    """Similar to `operator.itemgetter` except that it always expects and returns iterables.
    Compare `mapmap`
    """

    return lambda d: (d[i] for i in it)


def subdictdefault(d, it, default=None):
    # type: (Mapping[T, U], Iterable[T], Union[U, V]) -> Dict[T, Union[U, V]]

    """Uses the elements of `it` as keys to extract a new sub-dictionary."""

    return {key: d.get(key, default) for key in it}


def update(d1: dict, d2: dict) -> None:

    """Same as `dict.update` except that `None` values are skipped."""

    for k, v in d2.items():
        if v is not None:
            d1[k] = v


class keydefaultdict(defaultdict):

    """defaultdict which passes a key to the default factory."""

    def __missing__(self, key):
        # type: (Hashable, ) -> Any

        if self.default_factory is None:
            raise KeyError(key)
        else:
            value = self[key] = self.default_factory(key)
            return value


class KeyExistsError(KeyError):
    pass


class NoOverwriteDict(UserDict):

    """Dictionary which does not allow overwriting existing items."""

    def __setitem__(self, key, value):
        if key in self.data:
            raise KeyExistsError(repr(key))
        self.data[key] = value

    def overwrite(self, key, value):
        self.data[key] = value



from __future__ import generator_stop

import json
from typing import TYPE_CHECKING

from .http import URLRequest
from .ops import logical_xor

if TYPE_CHECKING:
    from typing import Callable, List, Optional, Tuple


class StreamWatcher:
    def __init__(self, api):
        self.api = api
        self.followed_names = api.get_followed()
        self.followed_online = {userid: False for userid in self.followed_names.keys()}

    def watch(self, notify_started, notify_stopped):
        # type: (Callable[[str, str, Optional[str]], None], Callable[[str, str], None]) -> None

        user_ids = self.followed_names.keys()
        streams = self.api.get_streams(user_ids)
        online = streams.keys()
        offline = user_ids - online

        for user_id in online:
            if not self.followed_online[user_id]:
                self.followed_online[user_id] = True
                notify_started(user_id, self.followed_names[user_id], streams[user_id].get("title", None))

        for user_id in offline:
            if self.followed_online[user_id]:
                self.followed_online[user_id] = False
                notify_stopped(user_id, self.followed_names[user_id])


class TwitchAPI:

    base = "https://api.twitch.tv/helix/"
    login = "users"
    follows = "users/follows"
    streams = "streams?user_id={}"

    def __init__(self, client_id, userid=None, username=None):
        # type: (str, Optional[str], Optional[str]) -> None

        if not logical_xor(userid, username):
            raise ValueError("Either the userid or the username must be given")

        self.client_id = client_id

        if userid:
            self.userid = userid
        else:
            assert username
            self.userid = self.get_userid(username)

    def req(self, url, params):
        # type: (str, List[Tuple[str, str]]) -> dict

        qs = "&".join(k + "=" + v for k, v in params)
        data = URLRequest(url + "?" + qs, headers={"Client-ID": self.client_id}).load()
        return json.loads(data)

    def get_userid(self, username):
        # type: (str, ) -> str

        d = self.req(self.base + self.login, [("login", username)])
        return d["data"][0]["id"]

    def get_followed(self):
        d = self.req(self.base + self.follows, [("from_id", self.userid)])
        return {follow["to_id"]: follow["to_name"] for follow in d["data"]}

    def get_streams(self, user_ids):
        d = self.req(self.base + self.streams, [("user_id", user_id) for user_id in user_ids])
        ret = {stream["user_id"]: stream for stream in d["data"]}
        assert len(d["data"]) == len(ret), "More than one stream per user"
        return ret

    def watcher(self):
        # type: () -> StreamWatcher

        return StreamWatcher(self)



from __future__ import generator_stop

from typing import TYPE_CHECKING

from .exceptions import InconsistentState, NoResult

if TYPE_CHECKING:
    from pymongo.collection import Collection


def findone(db, query):
    # type: (Collection, dict) -> dict

    docs = list(db.find(query))

    if len(docs) == 0:
        raise NoResult("No result found")
    elif len(docs) == 1:
        return docs[0]
    else:
        raise InconsistentState("More than one result found")



from __future__ import generator_stop

from typing import TYPE_CHECKING

import toml

if TYPE_CHECKING:
    from typing import Any


def read_toml(path):
    # type: (str, ) -> Any

    with open(path, encoding="utf-8") as fr:
        return toml.load(fr)


def write_toml(obj, path):
    # type: (Any, str) -> None

    with open(path, "w", encoding="utf-8") as fw:
        toml.dump(obj, fw)



from __future__ import generator_stop

from enum import Enum
from typing import TYPE_CHECKING

from .func import identity

if TYPE_CHECKING:
    from typing import Callable, Optional, Sequence, TypeVar

    T = TypeVar("T")
    U = TypeVar("U")


class BisectRetVal(Enum):
    Lower = -1
    Equal = 0
    Higher = 1


def make_binary_search(target):
    # type: (int, ) -> Callable[[int], BisectRetVal]

    def binary_search_func(x):
        if x < target:
            return BisectRetVal.Higher
        elif x > target:
            return BisectRetVal.Lower
        else:
            return BisectRetVal.Equal

    return binary_search_func


def make_binary_search_sequence(seq, target, key=None):
    # type: (Sequence[T], U, Optional[Callable[[T], U]]) -> Callable[[int], BisectRetVal]

    _key = key or identity

    def binary_search_func(x):
        # type: (int, ) -> BisectRetVal

        result = _key(seq[x])
        if result < target:
            return BisectRetVal.Higher
        elif result > target:
            return BisectRetVal.Lower
        else:
            return BisectRetVal.Equal

    return binary_search_func


def bisect_left_generic(lo, hi, func):
    # type: (int, int, Callable[[int], BisectRetVal]) -> int

    """Generic left bisection (binary search). Finds a specific `lo<=x<=hi` where
    `func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
    to `BisectRetVal.Lower`.
    """

    while lo < hi:
        mid = (lo + hi) // 2
        result = func(mid)
        if result == BisectRetVal.Higher:
            lo = mid + 1
        else:
            hi = mid
    return lo


def bisect_right_generic(lo, hi, func):
    # type: (int, int, Callable[[int], BisectRetVal]) -> int

    """Generic right bisection (binary search). Finds a specific `lo<=x<=hi` where
    `func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
    to `BisectRetVal.Lower`.
    """

    while lo < hi:
        mid = (lo + hi) // 2
        result = func(mid)
        if result == BisectRetVal.Lower:
            hi = mid
        else:
            lo = mid + 1
    return lo


def bisect_left_sequence(seq, target, key=None):
    """see: bisect.bisect_left"""

    func = make_binary_search_sequence(seq, target, key)
    return bisect_left_generic(0, len(seq), func)


def bisect_right_sequence(seq, target, key=None):
    """see: bisect.bisect_right"""

    func = make_binary_search_sequence(seq, target, key)
    return bisect_right_generic(0, len(seq), func)


def search_sorted(seq, newitem, key=None):
    # type: (Sequence[T], U, Optional[Callable[[T], U]]) -> int

    _key = key or identity

    for i, item in enumerate(seq):
        if newitem < _key(item):
            return i

    return len(seq)


if __name__ == "__main__":
    seq = [1, 2, 3, 4, 4, 5]
    from bisect import bisect_left, bisect_right

    for target in (1, 2, 3, 4, 5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5):
        print(bisect_left_sequence(seq, target) == bisect_left(seq, target))
        print(search_sorted(seq, target) == bisect_right_sequence(seq, target) == bisect_right(seq, target))



from __future__ import generator_stop

import concurrent.futures
import logging
import signal
import threading
from collections import deque
from concurrent.futures._base import FINISHED
from multiprocessing import Pool
from queue import Empty, Full, Queue
from typing import Any, Callable, Generic, Iterable, Iterator, Optional, Sequence, Set, Tuple, TypeVar, Union

from .exceptions import NoResult, assert_choice

T = TypeVar("T")
U = TypeVar("U")

logger = logging.getLogger(__name__)


class Worker(threading.Thread):

    TASK_IDLE = 0
    TASK_COMPLETE = 1
    TASK_EXCEPTION = 2

    def __init__(
        self,
        tasks: "Queue[Optional[Tuple[int, int, Callable, tuple, dict]]]",
        returns: "Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]",
    ):

        threading.Thread.__init__(self)
        self.tasks = tasks
        self.returns = returns

        self.daemon = True
        self.running = True
        self.start()

    def run(self):
        # type: () -> None

        while self.running:
            # self.returns.put((self.TASK_IDLE, self))
            task = self.tasks.get(True)
            if task:
                state, id, func, args, kargs = task
            else:
                self.tasks.task_done()
                self.running = False
                break

            try:
                ret = func(*args, **kargs)
            except Exception as e:  # pylint: disable=broad-except
                logger.exception("threaded task failed")
                self.returns.put((state, self.TASK_EXCEPTION, (id, e)))
            else:
                self.returns.put((state, self.TASK_COMPLETE, (id, ret)))

            try:
                self.tasks.task_done()
            except ValueError:  # if the task was canceled inbetween
                pass

        logger.debug("thread ended")

    def cancel(self):
        # type: () -> None

        self.running = False


class ThreadPool:
    def __init__(self, num_threads):
        # type: (int, ) -> None

        self.num_threads = num_threads
        self.tasks: "Queue[Optional[Tuple[int, int, Callable, tuple, dict]]]" = Queue()
        self.returns: "Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]" = Queue()
        self.my_worker = [Worker(self.tasks, self.returns) for _ in range(num_threads)]
        self.state = 0  # this is used to be able to cancel (or ignore the results of ongoing) tasks

    def add_task(self, id: int, func: Callable[..., Any], *args: Any, **kargs: Any) -> None:

        self.tasks.put((self.state, id, func, args, kargs))

    def get(self) -> Tuple[int, Optional[Any]]:

        while True:  # so invalid states can be ignored
            state, type, result = self.returns.get(True)  # type: Tuple[int, int, Tuple[int, Union[Any, Exception]]]
            if state == self.state:
                if type == Worker.TASK_COMPLETE:
                    return result  # (id, ret)
                else:
                    id, e = result
                    return id, None  # maybe raise here?
            #  else: old result, ignore

    @staticmethod
    def clear(queue: "Queue[Any]") -> None:

        """Only undocumented methods."""

        with queue.mutex:
            queue.queue.clear()
            queue.all_tasks_done.notify_all()
            queue.unfinished_tasks = 0

    @staticmethod
    def _clear(queue: "Queue[Any]") -> None:

        """Only uses documented methods, but is slower."""

        while not queue.empty():
            queue.get()
            queue.task_done()

    def close(self, finishall: bool = False) -> None:

        """if `finishall` is True, it will finish all tasks currently in the queue,
        if `finishall` is False, only currently active tasks will be finished
        """

        if finishall:
            for i in range(self.num_threads):
                self.tasks.put(None)
        else:
            for w in self.my_worker:
                w.cancel()

    def cancel(self) -> None:

        """Does not cancel already running tasks. Just clears queue and ignores
        results of currently running tasks.
        """

        self.clear(self.tasks)
        self.clear(self.returns)
        self.state += 1

    def wait(self) -> None:

        self.tasks.join()


def gather_all_unsorted(
    threadpool: ThreadPool, func: Callable, params: Sequence, *args: Any, **kwargs: Any
) -> Iterator[Tuple[Any, Any]]:

    """Runs multiple tasks concurrently and returns all results in the order
    of execution finish as soon as possible.
    """

    for i, param in enumerate(params):
        threadpool.add_task(i, func, param, *args, **kwargs)

    for i in range(len(params)):
        yield threadpool.get()


def gather_any(threadpool: ThreadPool, func: Callable, params: Iterable, *args: Any, **kwargs: Any) -> Tuple[Any, Any]:

    """Runs multiple tasks concurrently and just returns the result of the first
    task which finishes.
    """

    for id, result in gather_all_unsorted(threadpool, func, params, *args, **kwargs):
        threadpool.cancel()
        return id, result  # result will be None, if tasked threw exception
    raise NoResult("No task returned a valid result")


def NotThreadSafe(verify=False):
    # type: (bool, ) -> Callable[[type], type]

    """Class decorator. Only works on new style classes (object)
    `verify=True` prohibits write access to attributes.
    """

    def OnClass(TheClass):
        # type: (type, ) -> type

        if verify:
            orig_init = TheClass.__init__
            orig_setattr = TheClass.__setattr__

            def __setattr__(self, name, value):
                if self.thread_ident == threading.current_thread().ident:
                    orig_setattr(self, name, value)
                else:
                    raise RuntimeError(f"{TheClass.__name__}.{name} = {value} called from different thread")

            def __init__(self, *args, **kws):
                orig_setattr(self, "thread_ident", threading.current_thread().ident)
                orig_init(self, *args, **kws)

            TheClass.__init__ = __init__
            TheClass.__setattr__ = __setattr__
        return TheClass

    return OnClass


class AbortIteration(Exception):
    pass


class IterWorker(threading.Thread):

    STATE_WAITING = 0
    STATE_RUNNING = 1
    STATE_PAUSING = 2
    STATE_PAUSED = 3
    STATE_STOPPED = 4

    CMD_RESUME = 0
    CMD_ABORT = 1

    def __init__(self, taskqueue: Queue, onstatechange: Optional[Callable] = None):

        threading.Thread.__init__(self)
        self.queue = taskqueue
        self.onstatechange = onstatechange
        self.control: "Queue[int]" = Queue()
        self.state = self.STATE_WAITING

    @property
    def state(self):
        # type: () -> int

        return self._state

    @state.setter
    def state(self, value):
        # type: (int, ) -> None

        assert_choice("value", value, (0, 1, 2, 3, 4))

        self._state = value
        if self.onstatechange:
            self.onstatechange(value)

    def run(self):
        # type: () -> None

        while True:
            iter = self.queue.get()
            if not iter:
                self.state = self.STATE_STOPPED
                return
            self.state = self.STATE_RUNNING
            try:
                for __ in iter:
                    if self.state == self.STATE_PAUSING:
                        self.state = self.STATE_PAUSED
                        cmd = self.control.get()
                        if cmd == self.CMD_RESUME:
                            self.state = self.STATE_RUNNING
                            continue
                        elif cmd == self.CMD_ABORT:
                            break
            except AbortIteration:
                logger.warning("Task aborted")
            except Exception:
                logger.exception("Task threw exception")
            self.queue.task_done()  # untested
            self.state = self.STATE_WAITING

    def pause(self):
        # type: () -> None

        if self.state == self.STATE_RUNNING:
            self.state = self.STATE_PAUSING

    def resume(self):
        # type: () -> None

        if self.state == self.STATE_PAUSED:
            self.control.put(self.CMD_RESUME)

    def stop(self):
        # type: () -> None

        if self.state < self.STATE_STOPPED:
            self.queue.put(None)

    def clear(self):
        # type: () -> None

        try:
            while True:
                self.queue.get_nowait()
                self.queue.task_done()
        except Empty:
            pass

    def abort(self):
        # type: () -> None

        self.pause()
        self.control.put(self.CMD_ABORT)


# was: DownloadWorker
class ProgressWorker(threading.Thread):
    def __init__(self, manager, tasks):
        threading.Thread.__init__(self)
        self.manager = manager
        self.tasks = tasks

        self.running = True
        self.task, self.done, self.total = None, None, None

    def run(self):
        while self.running:
            self.task = self.tasks.get()
            if self.task:
                func, args, kwargs = self.task
            else:
                break

            try:
                result = func(self.report, *args, **kwargs)
            except Exception:
                logger.exception("Task failed")
                result = None
            self.task, self.done, self.total = None, None, None
            self.manager.finalize(result)

    def quit(self):
        self.running = False

    def report(self, done, total):
        self.done = done
        self.total = total

    def get(self):  # not using locks atm, can return inconsistent data
        if self.task and self.done and self.total:
            return (self.task, self.done, self.total)
        else:
            return None


# was: ThreadedDownloader
class ProgressThreadPool:
    def __init__(self, concurrent=1):
        # from config
        self.concurrent = concurrent

        self.waiting_queue = Queue()
        self.completed = []
        self.failed = []
        self.lock = threading.Lock()

        self.workers = list(ProgressWorker(self, self.waiting_queue) for i in range(self.concurrent))
        for w in self.workers:
            w.daemon = True  # program will end even if threads are still running
            w.start()

    def start(self, callable, *args, **kwargs):  # todo: add optional task id here.
        self.waiting_queue.put((callable, args, kwargs))

    def finalize(self, result):
        if result:
            setter, status, localname, length = result
            with self.lock:
                if status:
                    self.failed.append((status, localname, length))
                else:
                    self.completed.append((localname, length))
                    setter(localname)
        else:  # task failed
            pass

    def clear_completed(self):
        self.completed = []  # assignment is atomic right?

    def cancel_pending(self):
        # copied from https://stackoverflow.com/a/18873213
        with self.waiting_queue.mutex:
            self.waiting_queue.queue.clear()
            self.waiting_queue.all_tasks_done.notify_all()
            self.waiting_queue.unfinished_tasks = 0

    def get_waiting(self):
        return list(task for task in self.waiting_queue.queue if task)

    def get_completed(self):
        return self.completed

    def get_failed(self):
        return self.failed

    def get_running(self):  # no locks, inconsistent data
        return list(task for task in (w.get() for w in self.workers) if task)


class BoundedQueue:

    # similar: pip install bounded-iterator

    """Semaphor bounded queue. Can be used with `multiprocessing.Pool` for example.
    `imap()` calls iterable from same process but different thread.

    Other semaphore classes like `multiprocessing.BoundedSemaphore` can be used as well.
    """

    def __init__(self, size, it, timeout=None, semaphore=threading.BoundedSemaphore):
        # type: (int, Iterable[T], Optional[float], Callable[[int], Any]) -> None

        self.semaphore = semaphore(size)
        self.iterable = it
        self.timeout = timeout
        self.iterator = None  # type: Optional[Iterator[T]]

    def __iter__(self):
        # type: () -> Iterator[T]

        self.iterator = iter(self.iterable)
        return self

    def __next__(self):
        # type: () -> T

        if self.iterator is None:
            raise TypeError

        if not self.semaphore.acquire(timeout=self.timeout):
            raise Full("BoundedQueue semaphore acquisition timed out")

        return next(self.iterator)

    def done(self):
        # type: () -> None

        self.semaphore.release()


class BufferedIterable(Generic[T]):
    def __init__(self, it: Iterable[T], bufsize: int):
        self.iterable = it
        self.iterator = None
        self.buffer = deque([])
        self.bufsize = bufsize

    def __iter__(self) -> Iterator[T]:

        self.iterator = iter(self.iterable)
        return self

    def __next__(self) -> T:

        if self.iterator is None:
            raise TypeError

        try:
            while len(self.buffer) < self.bufsize:
                item = next(self.iterator)
                self.buffer.append(item)
        except StopIteration:
            pass

        try:
            return self.buffer.popleft()
        except IndexError:
            raise StopIteration


class CompletedFutures:
    def __init__(self, it: Iterable[concurrent.futures.Future], bufsize: int, timeout=None):
        self.iterable = it
        self.iterator: Optional[Iterator[concurrent.futures.Future]] = None
        self.futures: Set[concurrent.futures.Future] = set()
        self.bufsize = bufsize
        self.timeout = timeout

    def __iter__(self) -> Iterator[concurrent.futures.Future]:

        self.iterator = iter(self.iterable)
        return self

    def __next__(self) -> concurrent.futures.Future:

        if self.iterator is None:
            raise TypeError

        try:
            while len(self.futures) < self.bufsize:
                item = next(self.iterator)
                self.futures.add(item)
        except StopIteration:
            pass

        done, not_done = concurrent.futures.wait(
            self.futures, timeout=self.timeout, return_when=concurrent.futures.FIRST_COMPLETED
        )
        if done:
            ret = done.pop()
            self.futures = done | not_done
            return ret

        raise StopIteration

    def cancel_pending(self):
        for future in self.futures:
            future.cancel()


def _ignore_sigint():
    # type: () -> None

    """This need to be pickle'able to work with `multiprocessing.Pool`."""

    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    except ValueError:  # ignore for threadpools as 'signal only works in main thread'
        pass


def parallel_map(
    func: Callable[[T], U],
    it: Iterable[T],
    poolcls: Optional[Callable] = None,
    ordered: bool = True,
    parallel: bool = True,
    workers: Optional[int] = None,
    bufsize: int = 1,
    chunksize: int = 1,
) -> Iterator[U]:

    """Parallel map which uses multiprocessing to distribute tasks.
    `bufsize` should be used to limit memory usage when the iterable `it`
    can be processed faster than the output iterator is consumed.

    Keyboard interrupts are ignored in child processes. They should however be terminated
    correctly by the main thread.
    """

    if parallel:

        if poolcls is None:
            poolcls = Pool

        q = BoundedQueue(bufsize, it)

        with poolcls(workers, _ignore_sigint) as p:

            if ordered:
                process = p.imap
            else:
                process = p.imap_unordered

            try:
                for item in process(func, q, chunksize):
                    yield item
                    q.done()
            except GeneratorExit:
                logging.warning("interrupted")
                q.done()  # the semaphore in the bounded queue might block otherwise
                raise
    else:
        yield from map(func, it)


def FutureWithResult(result):
    future = concurrent.futures.Future()
    future._result = result
    future._state = FINISHED
    return future


def executor_map(
    func: Callable[[T], U],
    it: Iterable[T],
    executercls: Optional[Callable] = None,
    ordered: bool = True,
    parallel: bool = True,
    workers: Optional[int] = None,
    bufsize: int = 1,
) -> Iterator[concurrent.futures.Future]:

    """Starts processing when the iterator is started to be consumed."""

    if parallel:

        if executercls is None:
            executercls = concurrent.futures.ThreadPoolExecutor

        def futures() -> Iterator[concurrent.futures.Future]:
            for item in it:
                yield executor.submit(func, item)

        with executercls(workers) as executor:
            if ordered:
                bufit: Iterable = BufferedIterable(futures(), bufsize + executor._max_workers)
            else:
                bufit = CompletedFutures(futures(), bufsize + executor._max_workers)

            yield from bufit

    else:
        yield from map(FutureWithResult, map(func, it))


class ThreadsafeList(list):  # untested!!!

    """This is a list object with context manager to handle locking to perform multiple operations
    on a list in a threadsafe way.
    Example.
    l = ThreadsafeList([1, 2, 3])
    l.append(4) # atomic, no locking needed
    with l:
            l[0] += 4 # not atomic, thus executed with context manager
    """

    def __init__(self, it):
        list.__init__(self, it)
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()


def idsleeprandom(i):
    import random
    import time

    time.sleep(random.random())  # nosec
    return i


def benchmark():
    """According to these benchmark results, executor_map is faster for threads,
    whereas parallel_map is faster for processes.
    """

    from multiprocessing.pool import ThreadPool

    from genutility.func import identity
    from genutility.iter import consume
    from genutility.time import PrintStatementTime

    executercls = concurrent.futures.ThreadPoolExecutor
    with PrintStatementTime("executor_map(ThreadPool) ordered=True took {delta}s"):
        consume(executor_map(identity, range(100000), executercls=executercls, ordered=True, parallel=True))
    with PrintStatementTime("executor_map(ThreadPool) ordered=False took {delta}s"):
        consume(executor_map(identity, range(100000), executercls=executercls, ordered=False, parallel=True))

    executercls = concurrent.futures.ProcessPoolExecutor
    with PrintStatementTime("executor_map(ProcessPool) ordered=True took {delta}s"):
        consume(executor_map(identity, range(10000), executercls=executercls, ordered=True, parallel=True))
    with PrintStatementTime("executor_map(ProcessPool) ordered=False took {delta}s"):
        consume(executor_map(identity, range(10000), executercls=executercls, ordered=False, parallel=True))

    poolcls = ThreadPool
    with PrintStatementTime("parallel_map(ThreadPool) ordered=True took {delta}s"):
        consume(parallel_map(identity, range(100000), poolcls=poolcls, ordered=True, parallel=True))
    with PrintStatementTime("parallel_map(ThreadPool) ordered=False took {delta}s"):
        consume(parallel_map(identity, range(100000), poolcls=poolcls, ordered=False, parallel=True))

    poolcls = Pool
    with PrintStatementTime("parallel_map(ProcessPool) ordered=True took {delta}s"):
        consume(parallel_map(identity, range(10000), poolcls=poolcls, ordered=True, parallel=True))
    with PrintStatementTime("parallel_map(ProcessPool) ordered=False took {delta}s"):
        consume(parallel_map(identity, range(10000), poolcls=poolcls, ordered=False, parallel=True))


if __name__ == "__main__":
    benchmark()



from __future__ import generator_stop

import logging
import os
import subprocess  # nosec
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from ..os import CurrentWorkingDirectory
from ..string import surrounding_join

logger = logging.getLogger(__name__)


class RarError(Exception):
    def __init__(self, msg, returncode=None, cmd=None, output=""):
        Exception.__init__(self, msg)
        self.returncode = returncode
        self.cmd = cmd
        self.output = output


"""WinRAR Exit Codes
255 	USER BREAK	User stopped the process
9	CREATE ERROR	Create file error
8	MEMORY ERROR	Not enough memory for operation
7	USER ERROR	Command line option error
6	OPEN ERROR	Open file error
5	WRITE ERROR	Write to disk error
4	LOCKED ARCHIVE	Attempt to modify an archive previously locked by the 'k' command
3	CRC ERROR	A CRC error occurred when unpacking
2	FATAL ERROR	A fatal error occurred
1	WARNING	Non fatal error(s) occurred
0	SUCCESS	Successful operation (User exit)"""


class Rar:

    windows_executable = Path("C:/Program Files/WinRAR/Rar.exe")

    def __init__(self, archive: Path, executable: Optional[Path] = None) -> None:

        """archive: archive to work with, everything which is supported by winrar
        executable: path/to/Rar.exe"""

        self.exe = executable or self.windows_executable

        if not self.exe.is_file():
            raise ValueError("Invalid executable")

        self.archive = archive
        self.filelist = []  # type: List[str]
        self.cmd = ""
        self.flags = {}  # type: Dict[str, Tuple[bool, str]]
        self.flags["lock"] = (False, "k")
        self.flags["delete"] = (False, "df")
        self.flags["test"] = (False, "t")
        self.flags["filetime"] = (False, "tl")
        self.flags["append_archive_name"] = (False, "ad")
        self.options = {}  # type: Dict[str, Tuple[bool, str]]
        self.options["split"] = (False, "v%s")
        self.options["compression"] = (False, "m%u")
        self.options["password"] = (False, "p%s")
        self.options["password_header"] = (False, "hp%s")
        self.options["recovery"] = (False, "rr%up")
        self.options["recovery_volumes"] = (False, "rr%u%%")

    def add_file(self, pathname):
        # type: (str, ) -> None

        self.filelist.append(pathname)

    def add_files(self, filelist):
        # type: (Iterable[str], ) -> None

        self.filelist.extend(filelist)

    def set_compression(self, level):
        """level: 0 store, 1 fastest, 2 fast, 3 normal, 4 good, 5 best (default: 3)"""

        if level not in range(0, 6) and level is not False:
            raise RarError("Invalid parameter: Set compression level (0-store...3-default...5-best)")
        self.options["compression"][0] = level

    def set_password(self, password, encrypt_filenames=False):
        if encrypt_filenames:
            self.options["password_header"][0] = password
            self.options["password"][0] = False
        else:
            self.options["password"][0] = password
            self.options["password_header"][0] = False

    def add_recovery_info(self, rr):
        """rr: recovery record in percent, only 1-10 is valid"""

        if rr < 1 or rr > 10:
            raise RarError("Only 1%-10% valid")
        self.options["recovery"][0] = rr

    def add_recovery_volumes(self, rv):
        """rv: number of recovery volumes"""

        self.options["recovery_volumes"][0] = rv

    def split(self, split):
        self.options["split"][0] = split

    def lock(self, flag=True):
        self.flags["lock"][0] = flag

    def delete_after_archiving(self, flag=True):
        """flag (bool): delete files after archiving"""

        self.flags["delete"][0] = flag

    def test_after_archiving(self, flag=True):
        """flag (bool): test archive after archiving"""

        self.flags["test"][0] = flag

    def set_archive_to_filetime(self, flag=True):
        self.flags["filetime"][0] = flag

    def commandline(self, cmd):
        self.cmd = cmd.strip()

    def _execute(self, args):
        cmd = f"{self.exe} {args}"
        logger.debug("CMD: " + cmd)
        try:
            ret = subprocess.check_output(cmd, stderr=subprocess.STDOUT, cwd=os.getcwd())  # nosec
        except UnicodeEncodeError:
            raise RarError("UnicodeError, Win32Console fault")
        except subprocess.CalledProcessError as e:
            assert isinstance(e.output, str)
            raise RarError("Error", e.returncode, e.cmd, e.output)  # should use only stderr

        return ret

    def test(self, password="-"):  # nosec
        self._execute(f't -p{password} "{self.archive}"')

    def get_files_str(self):
        return surrounding_join(" ", self.filelist, '"', '"')

    def get_flag_str(self):
        return surrounding_join(" ", [value[1] for value in self.flags.itervalues() if value[0] is True], "-", "")

    def get_options_str(self):
        return surrounding_join(
            " ", [value[1] % value[0] for value in self.options.itervalues() if value[0] is not False], "-", ""
        )

    def _get_args(self, command):
        return f'{command} {self.get_flag_str()} {self.get_options_str()} {self.cmd} "{self.archive}" {self.get_files_str()}'

    def create(self):
        self._execute(self._get_args("a"))

    def extract(self, dir, mode=0):
        self.filelist = [dir]
        if mode == 1:
            self.flags["append_archive_name"][0] = True
        self._execute(self._get_args("x"))

    def close(self):
        pass


def create_rar_from_folder(
    path, dest_path=None, profile_setter_func=None, filter_func=lambda x: True, name_transform=lambda x: x
):
    # type: (Path, Optional[Path], Callable, Callable, Callable) -> bool

    if not path.is_dir():
        return False

    if dest_path is None:
        dest_path = path.parent

    with CurrentWorkingDirectory(path):

        try:
            r = Rar(dest_path / f"{name_transform(path.name)}.rar")
            if profile_setter_func:
                profile_setter_func(r)
            with os.scandir(".") as it:
                for entry in filter(filter_func, it):  # was: scandir_rec
                    r.add_file(entry.path)
            r.create()
        except RarError as e:
            logger.error(f"{str(e)}\n{e.output}")
            return False

    return True


def create_rar_from_file(path, dest_path=Path("."), profile_setter_func=None, name_transform=lambda x: x):
    # type: (Path, Path, Callable[[Rar], Any], Callable) -> bool

    if dest_path == ".":
        dest_path = path.parent

    with CurrentWorkingDirectory(path.parent):

        try:
            r = Rar(dest_path / f"{name_transform(path.name)}.rar")
            if profile_setter_func:
                profile_setter_func(r)
            r.add_file(path.name)
            r.create()
        except RarError as e:
            logger.error(f"{str(e)}\n{e.output}")
            return False

    return True



from __future__ import generator_stop

from ctypes import WinError
from msvcrt import get_osfhandle, open_osfhandle
from os import fdopen

from cwinsdk.um.handleapi import CloseHandle
from cwinsdk.um.winnt import GENERIC_READ, GENERIC_WRITE

_mode2access = {"": 0, "r": GENERIC_READ, "w": GENERIC_WRITE, "w+": GENERIC_READ | GENERIC_WRITE}


class WindowsHandle:
    def __init__(self, handle, doclose=True):
        # type: (int, bool) -> None

        if not isinstance(handle, int):
            raise ValueError("handle must be an int")

        self.handle = handle
        self.doclose = doclose

    @classmethod
    def from_file(cls, fp):
        return cls.from_fd(fp.fileno())

    @classmethod
    def from_fd(cls, fd):
        return cls(get_osfhandle(fd), doclose=False)

    def get_fd(self, flags):
        return open_osfhandle(self.handle, flags)

    def get_file(self, flags, mode="r", buffering=-1, encoding=None, errors=None, newline=None, closefd=True):
        return fdopen(self.get_fd(flags), mode, buffering, encoding, errors, newline, closefd)

    def close(self):
        # type: () -> None

        if CloseHandle(self.handle) == 0:
            raise WinError()

    def __enter__(self):
        # type: () -> WindowsHandle

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.doclose:
            self.close()



from __future__ import generator_stop

from collections import defaultdict
from functools import wraps
from itertools import product, zip_longest
from os import remove
from tempfile import NamedTemporaryFile
from time import sleep
from typing import TYPE_CHECKING
from unittest import TestCase
from unittest.result import TestResult
from unittest.runner import TextTestRunner

from .file import _check_arguments, equal_files
from .signal import HandleKeyboardInterrupt  # problem here

if TYPE_CHECKING:
    from typing import Any, Callable, Collection, DefaultDict, Hashable, Iterable, Mapping, Optional, Tuple, TypeVar

    T = TypeVar("T")
    U = TypeVar("U")


class nullcontext:  # see: .compat.contextlib
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


anullcontext = nullcontext()


class MyTestResult(TestResult):
    """Extension of TestResult to support numbering test cases"""

    def __init__(self, stream, descriptions, verbosity):
        """Initializes the test number generator, then calls super impl"""
        self.stream = stream
        self.current_test = 0
        self.current_subtest = 0
        return TestResult.__init__(self, stream, descriptions, verbosity)

    def progress(self):
        self.stream.write(f"Running test #{self.current_test}, subtest #{self.current_subtest}\r")

    def startTest(self, test):
        self.current_test += 1
        TestResult.startTest(self, test)
        self.progress()

    def stopTest(self, test):
        TestResult.stopTest(self, test)

    def stopTestRun(self):
        self.stream.write("\n")
        TestResult.stopTestRun(self)

    def addSubTest(self, test, subtest, outcome):
        """Called when a subtest finishes."""

        TestResult.addSubTest(self, test, subtest, outcome)
        self.testsRun += 1
        self.current_subtest += 1
        self.progress()


class MyTestRunner(TextTestRunner):
    """Extension of TestRunner to support numbering test cases"""

    resultclass = MyTestResult

    def run(self, test):
        """Stores the total count of test cases, then calls super impl"""

        self.total_test_cases = test.countTestCases()
        return TextTestRunner.run(self, test)

    def _makeResult(self):
        return TextTestRunner._makeResult(self)


class NoRaise:
    def __init__(self, testcase, message=None):
        # type: (TestCase, Optional[str]) -> None

        self.testcase = testcase
        self.message = message

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            if self.message:
                self.testcase.fail(self.message)
            else:
                self.testcase.fail(exc_value)


def make_comparable(d):
    # type: (Any, ) -> Any

    if isinstance(d, (str, int, type(None))):
        return d
    elif isinstance(d, list):
        return list(make_comparable(i) for i in d)
    elif isinstance(d, tuple):
        return tuple(make_comparable(i) for i in d)
    elif isinstance(d, dict):
        return make_comparable(list(d.items()))
    else:
        raise ValueError(f"must be list, tuple or dict, not {type(d)}")


def is_equal_unordered(seq_a, seq_b):
    # type: (Collection, Collection) -> bool

    if isinstance(seq_a, set) and isinstance(seq_b, set):
        return seq_a == seq_b

    seq_a = sorted(make_comparable(seq_a))
    seq_b = sorted(make_comparable(seq_b))
    return seq_a == seq_b


class MyTestCase(TestCase):
    def subTest(self, msg=None, **params):
        try:
            func = TestCase.subTest
        except AttributeError:
            return anullcontext
        else:
            return func(self, msg, **params)

    def assertAnd(self, first, second, msg=None):
        # type: (Any, Any, Optional[str]) -> None

        self.assertTrue(first and second, msg)

    def assertTypeEqual(self, first, second, msg=None):
        # type: (Any, Any, Optional[str]) -> None

        self.assertEqual(type(first), type(second), msg)

    def assertFilesEqual(self, first_path, second_path, msg=None):
        # type: (Any, Any, Optional[str]) -> None

        self.assertTrue(equal_files(first_path, second_path), msg)

    def assertIterEqual(self, first, second, msg=None):
        # type: (Iterable, Iterable, Optional[str]) -> None

        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertIterAlmostEqual(self, first, second, msg=None):
        # type: (Iterable, Iterable, Optional[str]) -> None

        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            try:
                self.assertAlmostEqual(a, b, msg=f"in iteration index {i}: {msg}")
            except TypeError:
                raise AssertionError("Invalid types (probably different length iters)")  # from None

    def assertAllEqual(self, args, msg=None):  # *args doesn't work in python2!?
        # type: (Iterable, Optional[str]) -> None

        it = iter(args)
        first = next(it)
        for second in it:
            self.assertEqual(first, second, msg)

    def assertIterIterEqual(self, first, second, msg=None):  # UNTESTED
        # type: (Iterable[Iterable], Iterable[Iterable], Optional[str]) -> None

        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertIterEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertPriorityEqual(self, first, second, msg=None):
        # type: (Iterable[Optional[Tuple[Hashable, Hashable]]], Iterable[Optional[Tuple[Hashable, Hashable]]], Optional[str]) -> None

        d1 = defaultdict(set)  # type: DefaultDict[Hashable, set]
        d2 = defaultdict(set)  # type: DefaultDict[Hashable, set]

        for a, b in zip_longest(first, second):
            if a is not None:
                val_1, p_1 = a
                if b is None and p_1 == 0:
                    continue

            if b is not None:
                val_2, p_2 = b
                if a is None and p_2 == 0:
                    continue

            self.assertAnd(a, b)

            if val_1 == 0 and val_2 == 0:
                continue

            d1[p_1].add(val_1)
            d2[p_2].add(val_2)
        self.assertEqual(d1, d2, msg)

    def assertUnorderedSeqEqual(self, first, second, msg=None):
        # type: (Iterable, Iterable, Optional[str]) -> None

        # fixme: use is_equal_unordered

        first = sorted(first)
        second = sorted(second)
        self.assertEqual(first, second, msg)

    def assertNoRaise(self, msg=None):
        # type: (Optional[str], ) -> NoRaise

        return NoRaise(self, msg)

    def assertEqualsOneOf(self, result, truths):
        # type: (T, Iterable[T]) -> None

        for truth in truths:
            if result == truth:
                return

        raise AssertionError("Result does not match one of the provided truths")  # from None

    def assertUnorderedMappingEqual(self, first, second, msg=None):
        # type: (Mapping[T, U], Mapping[T, U], Optional[str]) -> None

        self.assertEqual(len(first), len(second), msg)

        for a, b in zip(sorted(first.keys()), sorted(second.keys())):
            self.assertEqual(a, b)
            self.assertEqual(first[a], second[b])


def random_arguments(n, *funcs):
    # type: (int, *Callable[[], Any]) -> Callable[[Callable], Callable]

    def decorator(func):
        @wraps(func)
        def inner(self):
            for i in range(n):
                with self.subTest(str(i)):
                    if func(self, *(f() for f in funcs)) is not None:
                        raise AssertionError

        return inner

    return decorator


# also called: parameterize
def parametrize(*args_list):
    # # type: (*tuple, ) -> Callable[[Callable], Callable] # mypy error: syntax error in type comment

    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in args_list:
                with self.subTest(str(args)[:1000]):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def parametrize_product(*args_list):
    # # type: (*tuple, ) -> Callable[[Callable], Callable] # mypy error: syntax error in type comment

    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in product(*args_list):
                with self.subTest(str(args)):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def repeat(number):
    # type: (int, ) -> Callable[[Callable], Callable]

    def decorator(func):
        @wraps(func)
        def inner(self):
            for i in range(number):
                if func(self) is not None:  # no self.subTest(str(i))
                    raise AssertionError

        return inner

    return decorator


class closeable_tempfile:

    Uninterrupted = HandleKeyboardInterrupt()

    def __init__(self, mode="w+b", encoding=None):

        encoding = _check_arguments(mode, encoding)
        self.f = NamedTemporaryFile(mode=mode, encoding=encoding, delete=False)

    def __enter__(self):
        return self.f, self.f.name

    def __exit__2(self, type, value, traceback):
        with self.Uninterrupted:
            self.f.close()  # close in case user has not closed the file
            remove(self.f.name)

    def __exit__(self, type, value, traceback):
        try:
            if type == EOFError:
                sleep(1)  # wait for KeyboardInterrupt which might follow EOFError...
        finally:
            self.f.close()  # close in case user has not closed the file
            remove(self.f.name)



from __future__ import generator_stop

import time
from datetime import datetime, timedelta, timezone, tzinfo
from typing import TYPE_CHECKING, Optional, overload

if TYPE_CHECKING:
    from datetime import time as dtime

utcmin = datetime.min.replace(tzinfo=timezone.utc)
utcmax = datetime.max.replace(tzinfo=timezone.utc)


def is_aware(dt: datetime) -> bool:
    """Test if a datetime object is aware or naive."""

    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None

def naive_to_aware(dt, tzinfo=timezone.utc) -> datetime:
    if is_aware(dt):
        return dt
    else:
        return dt.replace(tzinfo=tzinfo)

def now(aslocal: bool = False) -> datetime:

    """Returns the current datetime as timezone aware object in
    UTC timezone if `aslocal=False` (the default)
    or local timezone if `aslocal=True`.
    """

    dt = datetime.now(timezone.utc)
    if aslocal:
        dt = dt.astimezone(None)
    return dt


def datetime_from_utc_timestamp(epoch: float, aslocal: bool = False) -> datetime:

    """Converts a UNIX epoch time in seconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    # don't use `fromtimestamp`, to have better cross platform support
    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=epoch)

    if aslocal:
        dt = dt.astimezone(None)
    return dt


def datetime_from_utc_timestamp_ms(epoch: int, aslocal: bool = False) -> datetime:

    """Converts a UNIX epoch time in milliseconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    seconds, ms = divmod(epoch, 1000)

    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=seconds, milliseconds=ms)

    if aslocal:
        dt = dt.astimezone(None)

    return dt


def datetime_from_utc_timestamp_ns(epoch: int, aslocal: bool = False) -> datetime:

    """Converts a UNIX epoch time in nano seconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    seconds, ns = divmod(epoch, 1000000000)

    # don't use `fromtimestamp`, to have better cross platform support
    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=seconds, microseconds=ns // 1000)

    if aslocal:
        dt = dt.astimezone(None)

    return dt


@overload
def between(dt, start, end):
    # type: (datetime, Optional[datetime], Optional[datetime]) -> bool

    pass


@overload
def between(dt, start, end):
    # type: (dtime, Optional[dtime], Optional[dtime]) -> bool

    pass


def between(dt, start=None, end=None):

    """Tests if `dt` is in-between `start` and `end` (inclusive and optionally open ended).

    If the parameters are datetimes, they all most be either offset-aware or native.
    If `start` and `end` are `time` objects (without dates),
    then `end` can come before `start` to specify ranges which overlap two days.

    If `start` equals `end`, return `True`.
    """

    if start and end:
        if start < end:
            return start <= dt and dt <= end
        else:
            return start <= dt or dt <= end

    if start and dt < start:
        return False
    if end and dt > end:
        return False
    return True


class LocalTimezone(tzinfo):
    def __init__(self):
        if time.daylight:
            self.daylight = timedelta(seconds=-time.altzone) - timedelta(seconds=-time.timezone)
            self.offset = timedelta(seconds=-time.altzone)
            self.name = time.tzname[1]
        else:
            self.daylight = timedelta(0)
            self.offset = timedelta(seconds=-time.timezone)
            self.name = time.tzname[0]

    def dst(self, dt):
        return self.daylight

    def utcoffset(self, dt):
        return self.offset

    def tzname(self):
        return self.name

    def __repr__(self):
        return "genutility.datetime.localtimezone"


localtimezone = LocalTimezone()

# deprecated
def localnow():
    # type: () -> datetime

    """Returns the current datetime as timezone aware object in local timezone."""

    return datetime.now(localtimezone)



from __future__ import generator_stop

import importlib
import logging
import os.path
import pickle  # nosec
from datetime import timedelta
from functools import wraps
from os import fspath
from pathlib import Path
from pickle import HIGHEST_PROTOCOL  # nosec
from typing import Any, Callable, ContextManager, Iterable, Iterator, Optional, Tuple

from .atomic import sopen
from .compat.contextlib import nullcontext
from .datetime import now
from .file import copen
from .filesystem import mdatetime
from .object import args_to_key
from .time import PrintStatementTime

logger = logging.getLogger(__name__)


def read_pickle(path: str) -> Any:

    """Read pickle file from `path`.
    Warning: All usual security consideration regarding the pickle module still apply.
    """

    with copen(path, "rb") as fr:
        return pickle.load(fr)  # nosec


def write_pickle(result: Any, path: str, protocol: Optional[int] = None, safe: bool = False) -> None:

    """Write `result` to `path` using pickle serialization.

    `protocol': pickle protocol version
    `safe`: if True, don't overwrite original file in case any error occurs
    """

    with sopen(path, "wb", safe=safe) as fw:
        pickle.dump(result, fw, protocol=protocol)


def read_iter(path: str) -> Iterator[Any]:

    """Read pickled iterable from `path`.
    Warning: All usual security consideration regarding the pickle module still apply.
    """

    with copen(path, "rb") as fr:
        unpickler = pickle.Unpickler(fr)  # nosec
        while fr.peek(1):
            yield unpickler.load()


def write_iter(it: Iterable[Any], path: str, protocol: Optional[int] = None, safe: bool = False) -> Iterator[Any]:

    """Write iterable `it` to `path` using pickle serialization. This uses much less memory than
            writing a full list at once.
    Read back using `read_iter()`. If `safe` is True, the original file is not overwritten
            if any error occurs.
    This is a generator which yields the values read from `it`. So it must be consumed
            to actually write anything to disk.
    """

    with sopen(path, "wb", safe=safe) as fw:
        pickler = pickle.Pickler(fw, protocol=protocol)
        for result in it:
            pickler.dump(result)
            yield result


def key_to_hash(key: Any, protocol: Optional[int] = None) -> str:
    from hashlib import md5

    binary = pickle.dumps(key, protocol=protocol)
    return md5(binary).hexdigest()  # nosec


def cache(
    path: Path,
    duration: Optional[timedelta] = None,
    generator: bool = False,
    protocol: int = HIGHEST_PROTOCOL,
    ignoreargs: bool = False,
    ignore_first_arg: bool = False,
    verbose: bool = False,
    keyfunc: Optional[Callable[[tuple, dict], str]] = None,
    consume: bool = False,
    return_cached: bool = False,
    file_ext: Optional[str] = None,
) -> Callable[[Callable], Callable]:

    """Decorator to cache function calls. Doesn't take function arguments into regard.
    It's using `pickle` to deserialize the data. So don't use it with untrusted inputs.

    `path`: path to cache file or directory.
            This can be a template string where `ppv` is assigned the pickle protocol version.
    `duration`: maximum age of cache
    `generator`: set to True to store the results of generator objects
    `protocol`: pickle protocol version
    `verbose`: print loading times if `generator` is False

    If `ignoreargs` is True, the cache won't take function arguments into regard.
            The path will be interpreted as template string to a file instead of a directory.
    """

    if not generator and consume:
        raise ValueError("consume can onlt be used for generator")

    if duration is None:
        _duration = timedelta.max
    else:
        _duration = duration

    if not verbose:
        context: Callable[[str], ContextManager] = nullcontext
    else:
        context = PrintStatementTime

    def decorator(func):
        # type: (Callable, ) -> Callable

        @wraps(func)
        def inner(*args, **kwargs):
            # type: (*Any, **Any) -> Any

            strpath = fspath(path).format(ppv=protocol)

            if not ignoreargs:
                if ignore_first_arg:
                    args = args[1:]

                _file_ext = file_ext or ".p"

                if keyfunc is None:
                    hashstr = key_to_hash(args_to_key(args, kwargs), protocol=protocol)
                else:
                    hashstr = keyfunc(args, kwargs)
                fullpath = os.path.join(strpath, hashstr + _file_ext)
            else:
                if keyfunc or file_ext:
                    raise ValueError("`keyfunc` or `file_ext` can only be specified if ignoreargs is False")
                if args or kwargs:
                    logger.warning("cache file decorator for %s called with arguments", func.__name__)
                fullpath = strpath

            try:
                invalid = now() - mdatetime(fullpath) > _duration
            except FileNotFoundError:
                invalid = True

            if invalid:
                cached = False
                if not ignoreargs:
                    path.mkdir(parents=True, exist_ok=True)

                if generator and not consume:
                    it = func(*args, **kwargs)
                    logger.info("Writing iterable to cache: %s", fullpath)
                    result: Any = write_iter(it, fullpath, protocol=protocol, safe=True)
                else:
                    with context("Result calculated in {delta} seconds"):
                        if generator and consume:
                            result = list(func(*args, **kwargs))
                        else:
                            result = func(*args, **kwargs)
                    logger.info("Writing result to cache: %s", fullpath)
                    write_pickle(result, fullpath, protocol=protocol, safe=True)
            else:
                cached = True
                if generator and not consume:
                    logger.info("Loading iterable from cache: %s", fullpath)
                    result = read_iter(fullpath)
                else:
                    with context(f"Result loaded from {fullpath} in {{delta}} seconds"):
                        result = read_pickle(fullpath)

            if return_cached:
                return cached, result
            else:
                return result

        return inner

    return decorator


def unpickle(path: str, requirements: Iterable[Tuple[str, Optional[str]]] = ()) -> Any:

    """Can be used to unpickle objects when normal unpickling fails to import some dependencies correctly.
    path: Path to the pickled file.
    requirements: Iterable of (module, package) tuples to be imported.

    Consider this scenario:
    ```python
    import pickle
    class asd:
            def qwe():
                    pass
    a = asd()
    with open("asd.p", "wb") as fw:
            pickle.dump(a, fw)

    with open("asd.p", "rb") as fr:
            a = pickle.load(fr)  # everything ok here
    ```

    Now the class definition is moved out of this file into another file `qwe.py`. Then
    ```python
    import pickle
    with open("asd.p", "rb") as fr:
            a = pickle.load(fr)  # AttributeError: Can't get attribute 'asd' on <module '__main__' (built-in)>
    ```

    Now unpickle can be used to fix this problem:
    ```python
    from genutility.pickle import unpickle
    a = unpickle("asd.p", [("qwe", "asd")])
    ```

    Warning: Only use on safe inputs.
    """

    import __main__

    for module, package in requirements:
        mod = importlib.import_module(module, package)
        if package:
            name = package
            type = getattr(mod, package)
        else:
            name = module
            type = mod

        setattr(__main__, name, type)

    with copen(path, "rb") as fr:
        return pickle.load(fr)  # nosec



from __future__ import generator_stop

from contextlib import ExitStack
from pathlib import Path
from typing import TYPE_CHECKING

from PyPDF2 import PdfFileMerger, PdfFileReader

if TYPE_CHECKING:
    from os import PathLike
    from typing import Iterator, Union


def join_pdfs_in_folder(path_in, file_out, overwrite=False):
    # type: (Path, Union[PathLike, str], bool) -> None

    """Join all PDFs in `path_in` and write to `file_out`."""

    merger = PdfFileMerger(strict=True)

    if not path_in.exists():
        raise FileNotFoundError(path_in)

    with ExitStack() as stack:
        for path in path_in.glob("*.pdf"):
            if path.is_file():
                fr = stack.enter_context(path.open("rb"))  # open does not support Path in <3.6
                merger.append(fr)

        if overwrite:
            mode = "w"
        else:
            mode = "x"

        with open(file_out, mode + "b") as fw:
            merger.write(fw)


def iter_pdf_text(path):
    # type: (str, ) -> Iterator[str]

    with open(path, "rb") as fr:
        pdf = PdfFileReader(fr)
        for i in range(pdf.getNumPages()):
            page = pdf.getPage(i)
            yield page.extractText()


def _read_pdf_pdfminer(path):
    from pdfminer.high_level import extract_text

    return extract_text(path)


def _read_pdf_tika(path):
    from tika import parser

    raw = parser.from_file(path)

    # return raw["metadata"]
    return raw["content"]


def read_pdf(path, engine="pdfminer"):
    # type: (str, str) -> str

    try:
        func = {
            "pdfminer": _read_pdf_pdfminer,
            "tika": _read_pdf_tika,
        }[engine]
    except KeyError:
        raise ValueError(f"Engine {engine} doesn't exist")

    return func(path)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Merge pdf files in directory into one file.")
    parser.add_argument("dir", help="input directory")
    parser.add_argument("out", help="output file path")
    args = parser.parse_args()

    join_pdfs_in_folder(args.dir, args.out)



from __future__ import generator_stop

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Set, TypeVar

    T = TypeVar("T")


def get(s):
    # type: (Set[T], ) -> T

    for i in s:
        return i

    raise KeyError("set empty")



from __future__ import generator_stop

import logging
from typing import TYPE_CHECKING

import cv2
import numpy as np

if TYPE_CHECKING:
    from typing import Iterator, Union

    import wx  # noqa: F401

logger = logging.getLogger(__name__)


def wx_to_cv_image(wximage):
    # type (wx.Image, ) -> np.ndarray

    buf = wximage.GetDataBuffer()
    arr = np.asarray(buf, dtype=np.uint8)

    return np.reshape(arr, (wximage.Height, wximage.Width, 3))[:, :, ::-1]  # RGB => BGR


def grayscale(cvimg):
    # type: (np.ndarray, ) -> np.ndarray

    return cv2.cvtColor(cvimg, cv2.COLOR_RGB2GRAY)


def iter_video(input=0, show=False):
    # type: (Union[str, int], bool) -> Iterator[np.ndarray]

    cap = cv2.VideoCapture(input)
    logger.debug("Reading video using %s backend", cap.getBackendName())

    try:
        while True:
            retval, image = cap.read()
            if retval:
                yield image
            else:
                break

            if show:
                cv2.imshow("iter_video", image)
                if cv2.waitKey(1) & 0xFF == ord("q"):
                    return
    finally:
        cap.release()
        if show:
            cv2.destroyAllWindows()



from __future__ import generator_stop

from operator import itemgetter
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple, TypeVar, Union

if TYPE_CHECKING:
    T = TypeVar("T")

    from graphviz import Digraph


class BKNode:

    __slots__ = ["value", "leaves"]

    def __init__(self, value, leaves):
        # type: (Any, Dict[Any, BKNode]) -> None

        self.value = value
        self.leaves = leaves

    def __str__(self):
        return str(self.value) + ": " + str(self.leaves)

    def __repr__(self):
        return repr(self.value) + ": " + repr(self.leaves)


class BKTree:

    """BK-tree. Used for nearest neighbor queries according to a discrete metric.
    Examples for the metric are the Manhattan distance or the Levenshtein distance.

    from Levenshtein import distance # pip install python-Levenshtein

    words = ["laptop", "security", "microsoft", "computer", "software", "tree", "algorithm", "desktop"]

    tree = BKTree(distance)
    tree.update(words)
    tree.saveimage("bk-tree.gv")
    """

    def __init__(self, distance_func):
        # type: (Callable[[Any, Any], int], ) -> None

        self.distance_func = distance_func
        self.root = None  # type: Optional[BKNode]

    def add(self, value):
        # type: (Any, ) -> None

        if self.root is None:
            self.root = BKNode(value, {})
            return

        node = self.root
        while True:
            distance = self.distance_func(value, node.value)
            try:
                node = node.leaves[distance]
            except KeyError:
                node.leaves[distance] = BKNode(value, {})
                break

    def update(self, values):
        # type: (Iterable[Any], ) -> None

        for value in values:
            self.add(value)

    def find(self, value, max_distance):
        # type: (Any, int) -> List[Tuple[int, Any]]

        """Returns all values from tree where the metric distance
        is less or equal to `max_distance`.
        """

        node = self.root
        ret = []  # type: List[Tuple[int, Any]]

        if node is None:
            return ret

        candidates = [node]  # is a deque better here?

        while candidates:
            candidate = candidates.pop()
            distance = self.distance_func(value, candidate.value)

            if distance <= max_distance:
                ret.append((distance, candidate.value))

            # instead of looking for candidates by searching,
            # one could also directly access the necessary keys in the dict
            for d, bknode in candidate.leaves.items():
                lower = distance - max_distance
                upper = distance + max_distance
                if lower <= d <= upper:
                    candidates.append(bknode)

        return ret

    @staticmethod
    def _find_by_distance(node, distance):
        # type: (BKNode, int) -> Iterator[Set[Any]]

        for d, bknode in node.leaves.items():
            if d == distance:
                nodeset = set(BKTree._values(bknode))
                nodeset.add(node.value)
                yield nodeset
            yield from BKTree._find_by_distance(bknode, distance)

    def find_by_distance(self, distance):
        # type: (int, ) -> Iterator[Set[Any]]

        """Find all sets of values base on `distance` between each other."""

        if self.root is None:
            return iter([])
        else:
            return BKTree._find_by_distance(self.root, distance)

    @staticmethod
    def _dot(dot, node):
        # type: (Digraph, BKNode) -> None

        for distance, childnode in node.leaves.items():
            dot.node(str(childnode.value))
            dot.edge(str(node.value), str(childnode.value), label=str(distance))
            BKTree._dot(dot, childnode)

    def saveimage(self, filename, format="png"):
        from graphviz import Digraph

        if self.root is None:
            raise ValueError("Tree is empty")

        dot = Digraph(format=format)
        dot.node(str(self.root.value))
        self._dot(dot, self.root)
        dot.render(filename)

    @staticmethod
    def _values(node):
        # type: (BKNode, ) -> Iterator[Any]

        yield node.value
        for leaf in node.leaves.values():
            yield from BKTree._values(leaf)

    def values(self):
        # type: () -> Iterator[Any]

        if self.root is None:
            return iter([])
        else:
            return self._values(self.root)

    def __iter__(self):
        # type: () -> Iterator[Any]

        return self.values()


class RoutingObject:
    def __init__(self, value):
        # type: (T, ) -> None

        self.value = value

        self.subtree = None  # type: Optional[NodeType]
        self.distance_to_parent = None  # type: Optional[float]
        self.covering_radius = None  # type: Optional[float]

    def __repr__(self):
        return f"RoutingObject({self.value}, radius={self.covering_radius}, tree={self.subtree})"

    def set_subtree(self, node):
        self.subtree = node
        node.parent_object = self


class LeafObject:
    def __init__(self, value):
        # type: (T, ) -> None

        self.value = value

        self.distance_to_parent = None  # type: Optional[float]

    def __repr__(self):
        return f"LeafObject({self.value})"

    def to_routing_object(self):
        return RoutingObject(self.value)


def isid(obj):
    if obj is None:
        return None
    else:
        return id(obj)


class MNode:

    maxsize = 4

    def __init__(self, distance_func, parent_node, parent_object):
        # type: (Callable, Optional[NodeType], Optional[ObjectType]) -> None

        self.distance_func = distance_func
        self.parent_node = parent_node
        self.parent_object = parent_object

    @property
    def is_root(self):
        # type: () -> bool

        return self.parent_node is None

    @property
    def is_full(self):
        # type: () -> bool

        return len(self.objects) >= self.maxsize

    def add(self, object):
        # type: (ObjectType, ) -> None

        self.objects.add(object)
        if isinstance(object, RoutingObject):
            object.subtree.parent_node = self

        if not self.is_root:
            object.distance_to_parent = self.distance_func(object.value, self.parent_object.value)

    def replace(self, o1, o2):
        # type: (ObjectType, ObjectType) -> None

        self.objects.remove(o1)
        self.objects.add(o2)

    def __repr__(self):
        return repr(
            {
                "parent_node": isid(self.parent_node),
                "parent_object": isid(self.parent_object),
                "objects": self.objects,
            }
        )


class InternalNode(MNode):
    def __init__(self, distance_func, parent_node, parent_object):
        # type: (Callable, Optional[NodeType], Optional[ObjectType]) -> None

        MNode.__init__(self, distance_func, parent_node, parent_object)
        self.objects = set()  # type: Set[RoutingObject]

    def set_objects(self, objects):
        # type: (Set[RoutingObject], ) -> None

        self.objects = objects
        op = self.parent_object

        for o in objects:
            assert o.subtree.parent_object is o
            o.subtree.parent_node = self

        if not self.is_root:
            for o in objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            if objects:
                op.covering_radius = max(o.distance_to_parent + o.covering_radius for o in objects)
            else:
                op.covering_radius = 0

    def update_objects(self):
        for o in self.objects:
            assert o.subtree.parent_object is o
            assert o.subtree.parent_node is self

        if not self.is_root:
            op = self.parent_object

            for o in self.objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            op.covering_radius = max(o.distance_to_parent + o.covering_radius for o in self.objects)


class LeafNode(MNode):
    def __init__(self, distance_func, parent_node, parent_object):
        # type: (Callable, Optional[NodeType], Optional[ObjectType]) -> None

        MNode.__init__(self, distance_func, parent_node, parent_object)
        self.objects = set()  # type: Set[LeafObject]

    def set_objects(self, objects):
        # type: (Set[LeafObject], ) -> None

        self.objects = objects
        op = self.parent_object

        if not self.is_root:
            for o in objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            if objects:
                op.covering_radius = max(o.distance_to_parent for o in objects)
            else:
                op.covering_radius = 0

    def update_objects(self):
        if not self.is_root:
            op = self.parent_object

            for o in self.objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            op.covering_radius = max(o.distance_to_parent for o in self.objects)


ObjectType = Union[RoutingObject, LeafObject]
NodeType = Union[InternalNode, LeafNode]

from random import sample


def _promote_random(distance_func, objects):
    # type: (Callable, Set[ObjectType]) -> Tuple[ObjectType, ObjectType]

    a, b = sample(objects, 2)
    return a, b


def _partition_generalized_hyperplane(distance_func, objects, o1, o2):
    # type: (Callable, Set[ObjectType], ObjectType, ObjectType) -> Tuple[Set[ObjectType], Set[ObjectType]]

    a = set()  # type: Set[ObjectType]
    b = set()  # type: Set[ObjectType]

    for o in objects:
        d1 = distance_func(o.value, o1.value)
        d2 = distance_func(o.value, o2.value)

        if d1 <= d2:
            a.add(o)
        else:
            b.add(o)

    return a, b


class MTree:

    """See: M-tree: An Efficient Access Method for Similarity Search in Metric Spaces (1997)"""

    def __init__(self, distance_func, promote=None, partition=None, does_not_work_yet=None):
        # type: (Callable[[ObjectType, ObjectType], float], Optional[str], Optional[str], Optional[str]) -> None

        if does_not_work_yet != "OK":
            raise RuntimeError("MTree is work in progress")

        self.distance_func = distance_func

        self.root = LeafNode(self.distance_func, None, None)

        self._promote = {
            "random": _promote_random,
        }[promote or "random"]
        self._partition = {
            "generalized-hyperplane": _partition_generalized_hyperplane,
        }[partition or "generalized-hyperplane"]

    def __repr__(self):
        return f"<MTree: {self.root!r}>"

    def _keys(self, node):
        if isinstance(node, InternalNode):
            for ro in node.objects:
                yield from self._keys(ro.subtree)
        else:
            for lo in node.objects:
                yield lo.value

    def _split(self, node, object):
        # type: (NodeType, ObjectType) -> None

        # if not node.is_root:
        op = node.parent_object
        np = node.parent_node

        n_new = type(node)(self.distance_func, np, op)  # same type like existing
        all_objects = node.objects | {object}
        o1, o2 = self._promote(self.distance_func, all_objects)
        os1, os2 = self._partition(self.distance_func, all_objects, o1, o2)

        if isinstance(node, LeafNode):
            o1 = o1.to_routing_object()
            o2 = o2.to_routing_object()

        o1.set_subtree(node)
        o2.set_subtree(n_new)

        if node.is_root:
            # print("is_root")
            self.root = InternalNode(self.distance_func, None, None)
            self.root.add(o1)
            self.root.add(o2)

            node.set_objects(os1)  # recalculates distances
            n_new.set_objects(os2)

        else:
            np.replace(op, o1)
            if np.is_full:
                # print("is_full")
                self._split(np, o2)
            else:
                # print("is not full")
                np.add(o2)
                np.update_objects()

    def _add(self, node, object):
        # type: (NodeType, LeafObject) -> None

        if not isinstance(node, LeafNode):
            # print("not leaf", node)
            distances = [self.distance_func(ro.value, object.value) for ro in node.objects]

            n_in = [(ro, d) for ro, d in zip(node.objects, distances) if d <= ro.covering_radius]
            if n_in:
                found = min(n_in, key=itemgetter(1))[0]
            else:
                it = ((ro, d - ro.covering_radius, d) for ro, d in zip(node.objects, distances))
                min_entry = min(it, key=itemgetter(1))
                found = min_entry[0]
                found.covering_radius = min_entry[2]

            self._add(found.subtree, object)
        else:
            if not node.is_full:
                # print("not full", "LeafNode", node)
                node.add(object)
            else:
                # print("is full", "LeafNode", node)
                self._split(node, object)

    def _find(self, node, value, radius):

        assert not node.is_root

        distance_parent_to_value = self.distance_func(node.parent_object.value, value)

        if not isinstance(node, LeafNode):
            for ro in node.objects:
                if abs(distance_parent_to_value - ro.distance_to_parent) <= radius + ro.covering_radius:
                    if self.distance_func(ro.value, value) <= radius + ro.covering_radius:
                        for value in self._find(ro.subtree, value, radius):
                            yield value
        else:
            for lo in node.objects:
                if abs(distance_parent_to_value - lo.distance_to_parent) <= radius:
                    if self.distance_func(lo.value, value) <= radius:
                        yield lo.value

    def keys(self):
        yield from self._keys(self.root)

    def add(self, value):
        # type: (T, ) -> None

        self._add(self.root, LeafObject(value))

    def find(self, value, radius):
        # type: (T, float) -> Iterator[T]

        node = self.root

        if not isinstance(node, LeafNode):
            for ro in node.objects:
                if self.distance_func(ro.value, value) <= radius + ro.covering_radius:
                    for value in self._find(ro.subtree, value, radius):
                        yield value
        else:
            for lo in node.objects:
                if self.distance_func(lo.value, value) <= radius:
                    yield lo.value


def len_dist(s1, s2):
    return abs(len(s1) - len(s2))


def vals(start, end):
    # type: (int, int) -> Set[str]

    return {str(i) * i for i in range(start, end + 1)}


from unittest import TestCase


class MTreeTests(TestCase):
    def test_mtree_one_layer(self):
        values = vals(1, 4)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.find("333", 1)) == vals(2, 4)
        assert set(mt.keys()) == values

    def test_mtree_two_layer(self):
        values = vals(1, 5)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.keys()) == values

    def test_mtree_three_layer(self):
        values = vals(1, 9)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.find("333", 1)) == vals(2, 4)
        assert set(mt.keys()) == values


if __name__ == "__main__":

    import unittest

    unittest.main()



from __future__ import generator_stop

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import bs4


def find_one(node, *args, **kwargs):
    # type: (bs4.element.Tag, *Any, **Any) -> bs4.element.Tag

    nodes = node.find_all(*args, **kwargs)

    if len(nodes) != 1:
        raise ValueError(f"Found {len(nodes)} nodes for: {args} {kwargs}")

    return nodes[0]
